"""Define playing indicators for simple events.

This submodules provides several classes to add specific musical
playing techniques to :class:`mutwo.events.basic.SimpleEvent` objects.
They mostly derive from traditional Western playing techniques and their
notation. Unlike indicators of the :mod:`mutwo.parameters.notation_indicators`
module, playing indicators have an effect on the played music and aren't
merely specifications of representation. The proper way to handle
playing  indicators should be via a :class:`PlayingIndicatorCollection`
object that should be attached to the respective :class:`SimpleEvent`.
The collection contains all possible playing indicators which are defined
in this module. :class:`mutwo.events.music.NoteLike` contain by default
a playing indicator collection.

There are basically two different types of playing indicators:

1, Playing indicators which can only be on or off (for instance
``bartok_pizzicato``, ``prall`` or ``laissez_vibrer``). They have
a :attr:`is_active` attribute which can either be :obj:`True`
or :obj:`False`.

2. Playing indicators with one or more arguments (for instance
:class:`Tremolo` with :attr:`n_flags` or :class:`Arpeggio` with
:attr:`direction`). Their :attr:`is_active` attribute can't be
set by the user and get automatically initialised depending on
if all necessary attributes are defined (then active) or
if any of the necessary attributes is set to :obj:`None` (then
not active).

**Example:**

Set playing indicators of :class:`NoteLike`:

>>> from mutwo.events import music
>>> my_note = music.NoteLike('c', 1 / 4, 'mf')
>>> my_note.playing_indicators.articulation.name = "."  # add staccato
>>> my_chord = music.NoteLike('c e g', 1 / 2, 'f')
>>> my_chord.playing_indicators.arpeggio.direction= "up"  # add arpeggio
>>> my_chord.playing_indicators.laissez_vibrer.is_active = True  # and laissez_vibrer

Attach :class:`PlayingIndicatorCollection` to :class:`SimpleEvent`:

>>> from mutwo.events import basic
>>> from mutwo.parameters import playing_indicators
>>> my_simple_event = basic.SimpleEvent()
>>> my_simple_event.playing_indicators = playing_indicators.PlayingIndicatorCollection()
"""

import dataclasses
import typing

from mutwo import parameters

__all__ = (
    "Tremolo",
    "Articulation",
    "Arpeggio",
    "Pedal",
    "StringContactPoint",
    "Hairpin",
    "Ornamentation",
    "ArtificalHarmonic",
    "Fermata",
    "PlayingIndicatorCollection",
)


@dataclasses.dataclass()
class Tremolo(parameters.abc.ImplicitPlayingIndicator):
    n_flags: typing.Optional[int] = None


@dataclasses.dataclass()
class Articulation(parameters.abc.ImplicitPlayingIndicator):
    name: typing.Optional[str] = None  # TODO(for future usage add typing.Literal)


@dataclasses.dataclass()
class Arpeggio(parameters.abc.ImplicitPlayingIndicator):
    direction: typing.Optional[str] = None  # TODO(for future usage add typing.Literal)


@dataclasses.dataclass()
class Pedal(parameters.abc.ImplicitPlayingIndicator):
    pedal_type: typing.Optional[str] = None  # TODO(for future usage add typing.Literal)
    pedal_activity: typing.Optional[bool] = True


@dataclasses.dataclass()
class StringContactPoint(parameters.abc.ImplicitPlayingIndicator):
    contact_point: typing.Optional[
        str
    ] = None  # TODO(for future usage add typing.Literal)


@dataclasses.dataclass()
class Ornamentation(parameters.abc.ImplicitPlayingIndicator):
    direction: typing.Optional[str] = None  # TODO(for future usage add typing.Literal)
    n_times: int = 1


@dataclasses.dataclass()
class ArtificalHarmonic(parameters.abc.ImplicitPlayingIndicator):
    n_semitones: typing.Optional[int] = None


@dataclasses.dataclass()
class Fermata(parameters.abc.ImplicitPlayingIndicator):
    fermata_type: typing.Optional[
        str
    ] = None  # TODO(for future usage add typing.Literal)


@dataclasses.dataclass()
class Hairpin(parameters.abc.ImplicitPlayingIndicator):
    # TODO(for future usage add typing.Literal['<', '>', '!')
    symbol: typing.Optional[str] = None


@dataclasses.dataclass(frozen=True)
class PlayingIndicatorCollection(
    parameters.abc.IndicatorCollection[parameters.abc.PlayingIndicator]
):
    # this is kind of redundant, but perhaps still better than without using
    # the `dataclasses` module
    articulation: Articulation = dataclasses.field(default_factory=Articulation)
    artifical_harmonic: ArtificalHarmonic = dataclasses.field(
        default_factory=ArtificalHarmonic
    )
    arpeggio: Arpeggio = dataclasses.field(default_factory=Arpeggio)
    bartok_pizzicato: parameters.abc.PlayingIndicator = dataclasses.field(
        default_factory=parameters.abc.ExplicitPlayingIndicator
    )
    fermata: Fermata = dataclasses.field(default_factory=Fermata)
    hairpin: Hairpin = dataclasses.field(default_factory=Hairpin)
    natural_harmonic: parameters.abc.PlayingIndicator = dataclasses.field(
        default_factory=parameters.abc.ExplicitPlayingIndicator
    )
    laissez_vibrer: parameters.abc.PlayingIndicator = dataclasses.field(
        default_factory=parameters.abc.ExplicitPlayingIndicator
    )
    ornamentation: Ornamentation = dataclasses.field(default_factory=Ornamentation)
    pedal: Pedal = dataclasses.field(default_factory=Pedal)
    prall: parameters.abc.PlayingIndicator = dataclasses.field(
        default_factory=parameters.abc.ExplicitPlayingIndicator
    )
    string_contact_point: StringContactPoint = dataclasses.field(
        default_factory=StringContactPoint
    )
    tie: parameters.abc.PlayingIndicator = dataclasses.field(
        default_factory=parameters.abc.ExplicitPlayingIndicator
    )
    tremolo: Tremolo = dataclasses.field(default_factory=Tremolo)



"""Constants to be used for and with :mod:`mutwo.converters.frontends.isis`.

The file mostly contains different flags for running ISiS.
The flag definitions are documented
`here <https://isis-documentation.readthedocs.io/en/latest/CmdLineArgs.html>`_.
"""

ISIS_PATH = "isis.sh"
"""The path to the ISiS shell script. When installing ISiS with the packed
'Install_ISiS_commandline.sh' script, the path should be 'isis.sh'."""


SILENT_FLAG = "--quiet"
"""Flag for preventing ISiS from printing any information
during rendering."""



"""Mutwo is an event based framework for composing music or other time-based arts.

The structure of the framework is centred around objects with a :attr:`duration`
property which are called **Events** (and which are defined in the :mod:`mutwo.events`
package). These events may have several more attributes besides their duration
attribute (e. g. pitch, volume). Abstractions for such attributes are defined in
:mod:`mutwo.parameters`. The :mod:`mutwo.converters` package aims to translate
mutwo data to third-party software data (or upside down). The :mod:`mutwo.generators`
package supports algorithmic generation of artistic data.
"""

from . import utilities
from . import parameters
from . import converters
from . import events
from . import generators



"""Render midi files (SMF) from mutwo data.

"""

import functools
import itertools
import operator
import typing
import warnings

import expenvelope  # type: ignore
import mido  # type: ignore

from mutwo.converters import abc
from mutwo.converters.frontends import midi_constants
from mutwo.converters import symmetrical

from mutwo import events
from mutwo import parameters
from mutwo import utilities

__all__ = ("MidiFileConverter",)

ConvertableEvents = typing.Union[
    events.basic.SimpleEvent,
    events.basic.SequentialEvent[events.basic.SimpleEvent],
    events.basic.SimultaneousEvent[
        events.basic.SequentialEvent[events.basic.SimpleEvent]
    ],
]


class MidiFileConverter(abc.Converter):
    """Class for rendering standard midi files (SMF) from mutwo data.

    Mutwo offers a wide range of options how the respective midi file shall
    be rendered and how mutwo data shall be translated. This is necessary due
    to the limited and not always unambiguous nature of musical encodings in
    midi files. In this way the user can tweak the conversion routine to her
    or his individual needs.

    :param path: where to write the midi file. The typical file type extension '.mid'
        is recommended, but not mandatory.
    :param simple_event_to_pitches: Function to extract from a
        :class:`mutwo.events.basic.SimpleEvent` a tuple that contains pitch objects
        (objects that inherit from :class:`mutwo.parameters.abc.Pitch`).
        By default it asks the Event for its :attr:`pitch_or_pitches` attribute
        (because by default :class:`mutwo.events.music.NoteLike` objects are expected).
        When using different Event classes than ``NoteLike`` with a different name for
        their pitch property, this argument should be overridden. If the function call
        raises an :obj:`AttributeError` (e.g. if no pitch can be extracted),
        mutwo will interpret the event as a rest.
    :param simple_event_to_volume: Function to extract the volume from a
        :class:`mutwo.events.basic.SimpleEvent` in the purpose of generating midi notes.
        The function should return an object that inhertis from
        :class:`mutwo.parameters.abc.Volume`. By default it asks the Event for
        its :attr:`volume` attribute (because by default
        :class:`mutwo.events.music.NoteLike` objects are expected).
        When using different Event classes than ``NoteLike`` with a
        different name for their volume property, this argument should be overridden.
        If the function call raises an :obj:`AttributeError` (e.g. if no volume can be
        extracted), mutwo will interpret the event as a rest.
    :param simple_event_to_control_messages: Function to generate midi control messages
        from a simple event. By default no control messages are generated. If the
        function call raises an AttributeError (e.g. if an expected control value isn't
        available) mutwo will interpret the event as a rest.
    :param midi_file_type: Can either be 0 (for one-track midi files) or 1 (for
         synchronous multi-track midi files). Mutwo doesn't offer support for generating
         type 2 midi files (midi files with asynchronous tracks).
    :param available_midi_channels: tuple containing integer where each integer
        represents the number of the used midi channel. Integer can range from 0 to 15.
        Higher numbers of available_midi_channels (like all 16) are recommended when
        rendering microtonal music. It shall be remarked that midi-channel 9 (or midi
        channel 10 when starting to count from 1) is often ignored by several software
        synthesizer, because this channel is reserved for percussion instruments.
    :param distribute_midi_channels: This parameter is only relevant if more than one
        SequentialEvent is passed to the convert method. If set to True each
        SequentialEvent only makes use of exactly n_midi_channel (see next parameter).
        If set to False each converted SequentialEvent is allowed to make use of all
        available channels. If set to True and the amount of necessary MidiTracks is
        higher than the amount of available channels, mutwo will silently cycle through
        the list of available midi channel.
    :param n_midi_channels_per_track: This parameter is only relevant for
        distribute_midi_channels == True. It sets how many midi channels are assigned
        to one SequentialEvent. If microtonal chords shall be played by
        one SequentialEvent (via pitch bending messages) a higher number than 1 is
        recommended. Defaults to 1.
    :param maximum_pitch_bend_deviation: sets the maximum pitch bending range in cents.
        This value depends on the particular used software synthesizer and its settings,
        because it is up to the respective synthesizer how to interpret the pitch
        bending messages. By default mutwo sets the value to 200 cents which
        seems to be the most common interpretation among different manufacturers.
    :param ticks_per_beat: Sets the timing precision of the midi file. From the mido
        documentation: "Typical values range from 96 to 480 but some use even more
        ticks per beat".
    :param instrument_name: Sets the midi instrument of all channels.
    :param tempo_envelope: All Midi files should specify their tempo. The default
        value of mutwo is 120 BPM (this is also the value that is assumed by any
        midi-file-reading-software if no tempo has been specified). Tempo changes
        are supported (and will be written to the resulting midi file).

    **Example**:

    >>> from mutwo.converters.frontends import midi
    >>> from mutwo.parameters import pitches
    >>> # midi file converter that assign a middle c to all events
    >>> midi_converter = midi.MidiFileConverter(
    >>>     'test.mid',
    >>>     simple_event_to_pitches=lambda event: (pitches.WesternPitch('c'),)
    >>> )

    **Disclaimer**:
        The current implementation doesn't support glissandi yet (only static pitches),
        time-signatures (the written time signature is always 4/4 for now) and
        dynamically changing tempo (ritardando or accelerando).
    """

    _tempo_point_converter = symmetrical.tempos.TempoPointConverter()

    def __init__(
        self,
        path: str,
        simple_event_to_pitches: typing.Callable[
            [events.basic.SimpleEvent], typing.Tuple[parameters.abc.Pitch, ...]
        ] = lambda event: event.pitch_or_pitches,  # type: ignore
        simple_event_to_volume: typing.Callable[
            [events.basic.SimpleEvent], parameters.abc.Volume
        ] = lambda event: event.volume,  # type: ignore
        simple_event_to_control_messages: typing.Callable[
            [events.basic.SimpleEvent], typing.Tuple[mido.Message, ...]
        ] = lambda event: tuple([]),
        midi_file_type: int = None,
        available_midi_channels: typing.Tuple[int, ...] = None,
        distribute_midi_channels: bool = False,
        n_midi_channels_per_track: int = None,
        maximum_pitch_bend_deviation: float = None,
        ticks_per_beat: int = None,
        instrument_name: str = None,
        tempo_envelope: expenvelope.Envelope = None,
    ):
        # TODO(find a less redundant way of setting default values)
        # set current default values if parameters aren't defined
        if midi_file_type is None:
            midi_file_type = midi_constants.DEFAULT_MIDI_FILE_TYPE

        if available_midi_channels is None:
            available_midi_channels = midi_constants.DEFAULT_AVAILABLE_MIDI_CHANNELS

        if n_midi_channels_per_track is None:
            n_midi_channels_per_track = midi_constants.DEFAULT_N_MIDI_CHANNELS_PER_TRACK

        if maximum_pitch_bend_deviation is None:
            maximum_pitch_bend_deviation = (
                midi_constants.DEFAULT_MAXIMUM_PITCH_BEND_DEVIATION_IN_CENTS
            )

        if ticks_per_beat is None:
            ticks_per_beat = midi_constants.DEFAULT_TICKS_PER_BEAT

        if instrument_name is None:
            instrument_name = midi_constants.DEFAULT_MIDI_INSTRUMENT_NAME

        if tempo_envelope is None:
            tempo_envelope = midi_constants.DEFAULT_TEMPO_ENVELOPE

        # check for correct values of midi specifications (have to be correct to be
        # able to write a readable midi file)
        self._assert_midi_file_type_has_correct_value(midi_file_type)
        self._assert_available_midi_channels_have_correct_value(available_midi_channels)

        # initialise the attributes of the class
        self.path = path

        self._simple_event_to_pitches = simple_event_to_pitches
        self._simple_event_to_volume = simple_event_to_volume
        self._simple_event_to_control_messages = simple_event_to_control_messages

        self._distribute_midi_channels = distribute_midi_channels
        self._n_midi_channels_per_track = n_midi_channels_per_track
        self._available_midi_channels = available_midi_channels
        self._midi_file_type = midi_file_type
        self._maximum_pitch_bend_deviation = maximum_pitch_bend_deviation
        self._ticks_per_beat = ticks_per_beat
        self._instrument_name = instrument_name

        self._tempo_envelope = tempo_envelope

        self._pitch_bending_warning = (
            "Maximum pitch bending is {} cents up or down!".format(
                maximum_pitch_bend_deviation
            )
        )

    # ###################################################################### #
    #                          static methods                                #
    # ###################################################################### #

    @staticmethod
    def _assert_midi_file_type_has_correct_value(midi_file_type: int):
        try:
            assert midi_file_type in (0, 1)
        except AssertionError:
            message = (
                "Unknown midi_file_type '{}'. Only midi type 0 and 1 are supported."
            )
            raise ValueError(message)

    @staticmethod
    def _assert_available_midi_channels_have_correct_value(
        available_midi_channels: typing.Tuple[int, ...],
    ):
        # check for correct range of each number
        for midi_channel in available_midi_channels:
            try:
                assert midi_channel in midi_constants.ALLOWED_MIDI_CHANNELS
            except AssertionError:
                message = (
                    "Found unknown midi channel '{}' in available_midi_channels."
                    .format(midi_constants.ALLOWED_MIDI_CHANNELS)
                )
                message += " Only midi channel '{}' are allowed.".format(
                    midi_constants.ALLOWED_MIDI_CHANNELS
                )
                raise ValueError(message)

        # check for duplicate
        try:
            assert len(available_midi_channels) == len(set(available_midi_channels))
        except AssertionError:
            message = "Found duplicate in available_midi_channels '{}'.".format(
                available_midi_channels
            )
            raise ValueError(message)

    @staticmethod
    def _adjust_beat_length_in_microseconds(
        tempo_point: typing.Union[
            utilities.constants.Real, parameters.tempos.TempoPoint
        ],
        beat_length_in_microseconds: int,
    ) -> int:
        """This method makes sure that ``beat_length_in_microseconds`` isn't too big.

        Standard midi files define a slowest allowed tempo which is around 3.5 BPM.
        In case the tempo is lower than this slowest allowed tempo, `mutwo` will
        automatically set the tempo to the lowest allowed tempo.
        """

        if beat_length_in_microseconds >= midi_constants.MAXIMUM_MICROSECONDS_PER_BEAT:
            beat_length_in_microseconds = midi_constants.MAXIMUM_MICROSECONDS_PER_BEAT
            message = "TempoPoint '{}' is too slow for Standard Midi Files. ".format(
                tempo_point
            )
            message += (
                "The slowest possible tempo is '{0}' BPM. Tempo has been set to"
                " '{0}' BPM.".format(
                    mido.tempo2bpm(midi_constants.MAXIMUM_MICROSECONDS_PER_BEAT)
                )
            )
            warnings.warn(message)

        return beat_length_in_microseconds

    # ###################################################################### #
    #                         helper methods                                 #
    # ###################################################################### #

    def _beats_per_minute_to_beat_length_in_microseconds(
        self, beats_per_minute: utilities.constants.Real
    ) -> int:
        """Method for converting beats per minute (BPM) to midi tempo.

        Midi tempo is stated in beat length in microseconds.
        """

        beat_length_in_seconds = self._tempo_point_converter.convert(beats_per_minute)
        beat_length_in_microseconds = int(
            beat_length_in_seconds * midi_constants.MIDI_TEMPO_FACTOR
        )
        return beat_length_in_microseconds

    def _find_available_midi_channels_per_sequential_event(
        self,
        simultaneous_event: events.basic.SimultaneousEvent[
            events.basic.SequentialEvent[events.basic.SimpleEvent]
        ],
    ) -> typing.Tuple[typing.Tuple[int, ...], ...]:
        """Find midi channels for each SequentialEvent.

        Depending on whether distribute_midi_channels has been set
        to True this method distributes all available midi channels
        on the respective SequentialEvents.
        """

        if self._distribute_midi_channels:
            available_midi_channels_cycle = itertools.cycle(
                self._available_midi_channels
            )
            available_midi_channels_per_sequential_event = tuple(
                tuple(
                    next(available_midi_channels_cycle)
                    for _ in range(self._n_midi_channels_per_track)
                )
                for _ in simultaneous_event
            )

        else:
            available_midi_channels_per_sequential_event = tuple(
                self._available_midi_channels for _ in simultaneous_event
            )

        return available_midi_channels_per_sequential_event

    def _beats_to_ticks(self, absolute_time: parameters.abc.DurationType) -> int:
        return int(self._ticks_per_beat * absolute_time)

    def _cent_deviation_to_pitch_bending_number(
        self, cent_deviation: utilities.constants.Real,
    ) -> int:
        pitch_bend_percent = (cent_deviation + self._maximum_pitch_bend_deviation) / (
            self._maximum_pitch_bend_deviation * 2
        )

        if pitch_bend_percent > 1:
            pitch_bend_percent = 1
            warnings.warn(self._pitch_bending_warning)

        if pitch_bend_percent < 0:
            pitch_bend_percent = 0
            warnings.warn(self._pitch_bending_warning)

        pitch_bending_number = round(
            midi_constants.MAXIMUM_PITCH_BEND * pitch_bend_percent
        )
        pitch_bending_number -= midi_constants.NEUTRAL_PITCH_BEND

        return pitch_bending_number

    def _tune_pitch(
        self, absolute_tick_start: int, pitch: parameters.abc.Pitch, midi_channel: int,
    ) -> typing.Tuple[int, mido.Message]:
        """This method finds midi data to represent the entered mutwo pitch."""

        frequency = pitch.frequency
        closest_midi_pitch = utilities.tools.find_closest_index(
            frequency, parameters.pitches_constants.MIDI_PITCH_FREQUENCIES
        )
        difference_in_cents_to_closest_midi_pitch = parameters.abc.Pitch.hertz_to_cents(
            parameters.pitches_constants.MIDI_PITCH_FREQUENCIES[closest_midi_pitch],
            frequency,
        )
        pitch_bending_number = self._cent_deviation_to_pitch_bending_number(
            difference_in_cents_to_closest_midi_pitch
        )

        if absolute_tick_start != 0:
            # if possible add bending one tick earlier to avoid glitches
            absolute_tick_start -= 1

        pitch_bending_message = mido.Message(
            "pitchwheel",
            channel=midi_channel,
            pitch=pitch_bending_number,
            time=absolute_tick_start,
        )

        return closest_midi_pitch, pitch_bending_message

    # ###################################################################### #
    #             methods for converting mutwo data to midi data             #
    # ###################################################################### #

    def _tempo_envelope_to_midi_messages(
        self, tempo_envelope: expenvelope.Envelope
    ) -> typing.Tuple[mido.MetaMessage, ...]:
        """Converts a SequentialEvent of ``EnvelopeEvent`` to midi Tempo messages."""

        absolute_times = utilities.tools.accumulate_from_zero(tempo_envelope.durations)

        midi_messages = []
        for absolute_time, tempo_point in zip(absolute_times, tempo_envelope.levels):
            absolute_tick = self._beats_to_ticks(absolute_time)
            beat_length_in_microseconds = self._beats_per_minute_to_beat_length_in_microseconds(
                tempo_point
            )

            beat_length_in_microseconds = MidiFileConverter._adjust_beat_length_in_microseconds(
                tempo_point, beat_length_in_microseconds
            )

            tempo_message = mido.MetaMessage(
                "set_tempo", tempo=beat_length_in_microseconds, time=absolute_tick
            )
            midi_messages.append(tempo_message)

        return tuple(midi_messages)

    def _note_information_to_midi_messages(
        self,
        absolute_tick_start: int,
        absolute_tick_end: int,
        velocity: int,
        pitch: parameters.abc.Pitch,
        available_midi_channels_cycle: typing.Iterator,
    ) -> typing.Tuple[mido.Message, ...]:
        """Generate 'pitch bending', 'note on' and 'note off' messages for one tone."""

        midi_channel = next(available_midi_channels_cycle)
        midi_pitch, pitch_bending_message = self._tune_pitch(
            absolute_tick_start, pitch, midi_channel
        )

        midi_messages = [pitch_bending_message]

        for time, message_name in (
            (absolute_tick_start, "note_on"),
            (absolute_tick_end, "note_off"),
        ):
            midi_messages.append(
                mido.Message(
                    message_name,
                    note=midi_pitch,
                    velocity=velocity,
                    time=time,
                    channel=midi_channel,
                )
            )

        return tuple(midi_messages)

    def _extracted_data_to_midi_messages(
        self,
        absolute_time: utilities.constants.Real,
        duration: parameters.abc.DurationType,
        available_midi_channels_cycle: typing.Iterator,
        pitch_or_pitches: typing.Tuple[parameters.abc.Pitch, ...],
        volume: parameters.abc.Volume,
        control_messages: typing.Tuple[mido.Message, ...],
    ) -> typing.Tuple[mido.Message, ...]:
        """Generates pitch-bend / note-on / note-off messages for each tone in a chord.

        Concatenates the midi messages for every played tone with the global control
        messages.

        Gets as an input relevant data for midi message generation that has been
        extracted from a :class:`mutwo.abc.Event` object.
        """

        absolute_tick_start = self._beats_to_ticks(absolute_time)
        absolute_tick_end = absolute_tick_start + self._beats_to_ticks(duration)
        velocity = volume.midi_velocity

        midi_messages = []

        # add control messages
        for control_message in control_messages:
            control_message.time = absolute_tick_start
            midi_messages.append(control_message)

        # add note related messages
        for pitch in pitch_or_pitches:
            midi_messages.extend(
                self._note_information_to_midi_messages(
                    absolute_tick_start,
                    absolute_tick_end,
                    velocity,
                    pitch,
                    available_midi_channels_cycle,
                )
            )

        return tuple(midi_messages)

    def _simple_event_to_midi_messages(
        self,
        simple_event: events.basic.SimpleEvent,
        absolute_time: utilities.constants.Real,
        available_midi_channels_cycle: typing.Iterator,
    ) -> typing.Tuple[mido.Message, ...]:
        """Converts ``SimpleEvent`` (or any object that inherits from ``SimpleEvent``).

        Return tuple filled with midi messages that represent the mutwo data in the
        midi format.
        """

        extracted_data = []

        # try to extract the relevant data
        is_rest = False
        for extraction_function in (
            self._simple_event_to_pitches,
            self._simple_event_to_volume,
            self._simple_event_to_control_messages,
        ):
            try:
                extracted_data.append(extraction_function(simple_event))
            except AttributeError:
                is_rest = True

        # if not all relevant data could be extracted, simply ignore the
        # event
        if is_rest:
            return tuple([])

        # otherwise generate midi messages from the extracted data
        midi_messages = self._extracted_data_to_midi_messages(
            absolute_time,
            simple_event.duration,
            available_midi_channels_cycle,
            *extracted_data  # type: ignore
        )
        return tuple(midi_messages)

    def _sequential_event_to_midi_messages(
        self,
        sequential_event: events.basic.SequentialEvent[events.basic.SimpleEvent],
        available_midi_channels: typing.Tuple[int, ...],
    ) -> typing.Tuple[mido.Message, ...]:
        """Iterates through the ``SequentialEvent`` and converts each ``SimpleEvent``.

        Return unsorted tuple of Midi messages where the time attribute of each message
        is the absolute time in ticks.
        """

        midi_data: typing.List[mido.Message] = []

        available_midi_channels_cycle = itertools.cycle(available_midi_channels)

        # fill midi track with the content of the sequential event
        for absolute_time, simple_event in zip(
            sequential_event.absolute_times, sequential_event
        ):
            midi_messages = self._simple_event_to_midi_messages(
                simple_event, absolute_time, available_midi_channels_cycle
            )
            midi_data.extend(midi_messages)

        return tuple(midi_data)

    def _midi_messages_to_midi_track(
        self,
        midi_data: typing.Tuple[mido.Message, ...],
        duration: parameters.abc.DurationType,
        is_first_track: bool = False,
    ) -> mido.MidiTrack:
        """Convert unsorted midi message with absolute timing to a midi track.

        In the resulting midi track the timing of the messages is relative.
        """

        # initialise midi track
        track = mido.MidiTrack([])
        track.append(mido.MetaMessage("instrument_name", name=self._instrument_name))

        if is_first_track:
            # standard time signature 4/4
            track.append(mido.MetaMessage("time_signature", numerator=4, denominator=4))
            midi_data += self._tempo_envelope_to_midi_messages(self._tempo_envelope)

        # sort midi data
        sorted_midi_data = sorted(midi_data, key=lambda message: message.time)

        # add end of track message
        duration_in_ticks = self._beats_to_ticks(duration)
        sorted_midi_data.append(
            mido.MetaMessage("end_of_track", time=duration_in_ticks)
        )

        # convert from absolute to relative time
        delta_ticks_per_message = tuple(
            message1.time - message0.time
            for message0, message1 in zip(sorted_midi_data, sorted_midi_data[1:])
        )
        delta_ticks_per_message = (sorted_midi_data[0].time,) + delta_ticks_per_message
        for dt, message in zip(delta_ticks_per_message, sorted_midi_data):
            message.time = dt

        # add midi data to midi track
        track.extend(sorted_midi_data)

        return track

    # ###################################################################### #
    #           methods for filling the midi file (only called once)         #
    # ###################################################################### #

    def _add_simple_event_to_midi_file(
        self, simple_event: events.basic.SimpleEvent, midi_file: mido.MidiFile
    ) -> None:
        """Adds simple event to midi file."""
        self._add_sequential_event_to_midi_file(
            events.basic.SequentialEvent([simple_event]), midi_file
        )

    def _add_sequential_event_to_midi_file(
        self,
        sequential_event: events.basic.SequentialEvent[events.basic.SimpleEvent],
        midi_file: mido.MidiFile,
    ) -> None:
        """Adds sequential event to midi file."""
        self._add_simultaneous_event_to_midi_file(
            events.basic.SimultaneousEvent([sequential_event]), midi_file
        )

    def _add_simultaneous_event_to_midi_file(
        self,
        simultaneous_event: events.basic.SimultaneousEvent[
            events.basic.SequentialEvent[events.basic.SimpleEvent]
        ],
        midi_file: mido.MidiFile,
    ) -> None:
        """Adds one simultaneous event to a midi file.

        Depending on the midi_file_type either adds a tuple of MidiTrack
        objects (for midi_file_type = 1) or adds only one MidiTrack
        (for midi_file_type = 0).
        """

        # TODO(split this method, make it more readable!)

        available_midi_channels_per_sequential_event = self._find_available_midi_channels_per_sequential_event(
            simultaneous_event
        )

        midi_data_per_sequential_event = tuple(
            self._sequential_event_to_midi_messages(
                sequential_event, available_midi_channels
            )
            for sequential_event, available_midi_channels in zip(
                simultaneous_event, available_midi_channels_per_sequential_event
            )
        )

        duration = simultaneous_event.duration

        # midi file type 0 -> only one track
        if self._midi_file_type == 0:
            midi_data_for_one_track = functools.reduce(
                operator.add, midi_data_per_sequential_event
            )
            midi_track = self._midi_messages_to_midi_track(
                midi_data_for_one_track, duration, is_first_track=True
            )
            midi_file.tracks.append(midi_track)

        # midi file type 1
        else:
            midi_tracks = (
                self._midi_messages_to_midi_track(
                    midi_data, duration, is_first_track=nth_midi_data == 0
                )
                for nth_midi_data, midi_data in enumerate(
                    midi_data_per_sequential_event
                )
            )
            midi_file.tracks.extend(midi_tracks)

    def _event_to_midi_file(self, event_to_convert: ConvertableEvents) -> mido.MidiFile:
        """Convert mutwo event object to mido MidiFile object."""

        midi_file = mido.MidiFile(
            ticks_per_beat=self._ticks_per_beat, type=self._midi_file_type
        )

        # depending on the event types timing structure different methods are called
        if isinstance(event_to_convert, events.basic.SimultaneousEvent):
            self._add_simultaneous_event_to_midi_file(event_to_convert, midi_file)
        elif isinstance(event_to_convert, events.basic.SequentialEvent):
            self._add_sequential_event_to_midi_file(event_to_convert, midi_file)
        elif isinstance(event_to_convert, events.basic.SimpleEvent):
            self._add_simple_event_to_midi_file(event_to_convert, midi_file)
        else:
            message = "Can't convert object '{}' of type '{}' to a MidiFile.".format(
                event_to_convert, type(event_to_convert)
            )
            message += " Supported types include all inherited classes "
            message += "from '{}'.".format(ConvertableEvents)
            raise TypeError(message)

        return midi_file

    # ###################################################################### #
    #               public methods for interaction with the user             #
    # ###################################################################### #

    def convert(self, event_to_convert: ConvertableEvents) -> None:
        """Render a Midi file to the converters path attribute from the given event.

        :param event_to_convert: The given event that shall be translated
            to a Midi file.

        The following example generates a midi file that contains a simple ascending
        pentatonic scale:

        >>> from mutwo.events import basic, music
        >>> from mutwo.parameters import pitches
        >>> from mutwo.converters.frontends import midi
        >>> ascending_scale = basic.SequentialEvent(
        >>>     [
        >>>         music.NoteLike(pitches.WesternPitch(pitch), duration=1, volume=0.5)
        >>>         for pitch in 'c d e g a'.split(' ')
        >>>     ]
        >>> )
        >>> midi_converter = midi.MidiFileConverter(
        >>>     'ascending_scale.mid', available_midi_channels=(0,)
        >>> )
        >>> midi_converter.convert(ascending_scale)

        **Disclaimer:** when passing nested structures, make sure that the
        nested object matches the expected type. Unlike other mutwo
        converter classes (like :class:`mutwo.converters.symmetrical.TempoConverter`)
        :class:`MidiFileConverter` can't convert infinitely nested structures
        (due to the particular way how Midi files are defined). The deepest potential
        structure is a :class:`mutwo.events.basic.SimultaneousEvent` (representing
        the complete MidiFile) that contains :class:`mutwo.events.basic.SequentialEvent`
        (where each ``SequentialEvent`` represents one MidiTrack) that contains
        :class:`mutwo.events.basic.SimpleEvent` (where each ``SimpleEvent``
        represents one midi note). If only one ``SequentialEvent`` is send,
        this ``SequentialEvent`` will be read as one MidiTrack in a MidiFile.
        If only one ``SimpleEvent`` get passed, this ``SimpleEvent`` will be
        interpreted as one MidiEvent (note_on and note_off) inside one
        MidiTrack inside one MidiFile.
        """

        midi_file = self._event_to_midi_file(event_to_convert)
        midi_file.save(filename=self.path)



"""Render signing signals from mutwo data via `ISiS <https://forum.ircam.fr/projects/detail/isis/>`_.

ISiS (IRCAM Singing Synthesis) is a `"command line application for singing
synthesis that can be used to generate singing signals by means of synthesizing
them from melody and lyrics."
<https://isis-documentation.readthedocs.io/en/latest/Intro.html#the-isis-command-line>`_.
"""

import os
import typing

from mutwo.events import basic

from mutwo import converters
from mutwo.converters.frontends import isis_constants
from mutwo import parameters
from mutwo.utilities import constants

__all__ = ("IsisScoreConverter", "IsisConverter")

ConvertableEvents = typing.Union[
    basic.SimpleEvent, basic.SequentialEvent[basic.SimpleEvent],
]
ExtractedData = typing.Tuple[
    # duration, consonants, vowel, pitch, volume
    parameters.abc.DurationType,
    typing.Tuple[str, ...],
    str,
    parameters.abc.Pitch,
    parameters.abc.Volume,
]


class IsisScoreConverter(converters.abc.EventConverter):
    """Class to convert mutwo events to a `ISiS score file. <https://isis-documentation.readthedocs.io/en/latest/score.html>`_

    :param path: where to write the ISiS score file
    :param simple_event_to_pitch: Function to extract an instance of
        :class:`mutwo.parameters.abc.Pitch` from a simple event.
    :param simple_event_to_volume:
    :param simple_event_to_vowel:
    :param simple_event_to_consonants:
    :param tempo: Tempo in beats per minute (BPM). Defaults to 60.
    :param global_transposition: global transposition in midi numbers. Defaults to 0.
    :param n_events_per_line: How many events the score shall contain per line.
        Defaults to 5.
    """

    # TODO(auto-tie rests (because ISiS doesn't understand more than one rests
    # between notes)

    def __init__(
        self,
        path: str,
        simple_event_to_pitch: typing.Callable[
            [basic.SimpleEvent], parameters.abc.Pitch
        ] = lambda simple_event: simple_event.pitch_or_pitches[0],  # type: ignore
        simple_event_to_volume: typing.Callable[
            [basic.SimpleEvent], parameters.abc.Volume
        ] = lambda simple_event: simple_event.volume,  # type: ignore
        simple_event_to_vowel: typing.Callable[
            [basic.SimpleEvent], str
        ] = lambda simple_event: simple_event.vowel,  # type: ignore
        simple_event_to_consonants: typing.Callable[
            [basic.SimpleEvent], typing.Tuple[str, ...]
        ] = lambda simple_event: simple_event.consonants,  # type: ignore
        tempo: constants.Real = 60,
        global_transposition: int = 0,
        default_sentence_loudness: typing.Union[constants.Real, None] = None,
        n_events_per_line: int = 5,
    ):
        self.path = path
        self._tempo = tempo
        self._global_transposition = global_transposition
        self._default_sentence_loudness = default_sentence_loudness
        self._n_events_per_line = n_events_per_line

        self._extraction_functions = (
            simple_event_to_consonants,
            simple_event_to_vowel,
            simple_event_to_pitch,
            simple_event_to_volume,
        )

    # ###################################################################### #
    #                           private methods                              #
    # ###################################################################### #

    def _make_key_from_extracted_data_per_event(
        self,
        key_name: str,
        extracted_data_per_event: typing.Tuple[ExtractedData],
        key: typing.Callable[[ExtractedData], typing.Tuple[str, ...]],
        seperate_with_comma: bool = True,
    ) -> str:
        objects_per_line: typing.List[typing.List[str]] = [[]]
        for nth_event, extracted_data in enumerate(extracted_data_per_event):
            objects_per_line[-1].extend(key(extracted_data))
            if nth_event % self._n_events_per_line == 0:
                objects_per_line.append([])

        object_join_string = ", " if seperate_with_comma else " "
        lines = [object_join_string.join(line) for line in objects_per_line if line]

        line_join_string = ",\n" if seperate_with_comma else "\n"
        line_join_string = "{}{}".format(line_join_string, " " * (len(key_name) + 2))

        return "{}: {}".format(key_name, line_join_string.join(lines))

    def _make_lyrics_section_from_extracted_data_per_event(
        self, extracted_data_per_event: typing.Tuple[ExtractedData],
    ) -> str:
        xsampa = self._make_key_from_extracted_data_per_event(
            "xsampa",
            extracted_data_per_event,
            lambda extracted_data: extracted_data[1] + (extracted_data[2],),
            False,
        )

        lyric_section = "[lyrics]\n{}".format(xsampa)
        return lyric_section

    def _make_score_section_from_extracted_data_per_event(
        self, extracted_data_per_event: typing.Tuple[ExtractedData],
    ) -> str:
        midi_notes = self._make_key_from_extracted_data_per_event(
            "midiNotes",
            extracted_data_per_event,
            lambda extracted_data: (str(extracted_data[3].midi_pitch_number),),
        )
        rhythm = self._make_key_from_extracted_data_per_event(
            "rhythm",
            extracted_data_per_event,
            lambda extracted_data: (str(extracted_data[0]),),
        )
        loud_accents = self._make_key_from_extracted_data_per_event(
            "loud_accents",
            extracted_data_per_event,
            lambda extracted_data: (str(extracted_data[4].amplitude),),
        )
        score_section = (
            "[score]\n{}\nglobalTransposition: {}\n{}\n{}\ntempo: {}".format(
                midi_notes,
                self._global_transposition,
                rhythm,
                loud_accents,
                self._tempo,
            )
        )
        return score_section

    def _convert_simple_event(
        self,
        simple_event_to_convert: basic.SimpleEvent,
        absolute_entry_delay: parameters.abc.DurationType,
    ) -> typing.Tuple[ExtractedData]:
        duration = simple_event_to_convert.duration

        extracted_data: typing.List[object] = [duration]
        for extraction_function in self._extraction_functions:
            try:
                extracted_information = extraction_function(simple_event_to_convert)
            except AttributeError:
                return (
                    (
                        duration,
                        tuple([]),
                        "_",
                        parameters.pitches.WesternPitch(
                            "c",
                            -1,
                            concert_pitch=440,
                            concert_pitch_octave=4,
                            concert_pitch_pitch_class=9,
                        ),
                        parameters.volumes.DirectVolume(0),
                    ),
                )

            extracted_data.append(extracted_information)

        return (tuple(extracted_data),)  # type: ignore

    def _convert_simultaneous_event(
        self,
        simultaneous_event_to_convert: basic.SimultaneousEvent,
        absolute_entry_delay: parameters.abc.DurationType,
    ):
        message = (
            "Can't convert instance of SimultaneousEvent to ISiS Score. ISiS is only a"
            " monophonic synthesizer and can't read multiple simultaneous voices!"
        )
        raise NotImplementedError(message)

    # ###################################################################### #
    #                             public api                                 #
    # ###################################################################### #

    def convert(self, event_to_convert: ConvertableEvents) -> None:
        """Render ISiS score file from the passed event.

        :param event_to_convert: The event that shall be rendered to a ISiS score
            file.

        **Example:**

        >>> from mutwo.events import basic, music
        >>> from mutwo.parameters import pitches
        >>> from mutwo.converters.frontends import isis
        >>> notes = basic.SequentialEvent(
        >>>    [
        >>>         music.NoteLike(pitches.WesternPitch(pitch_name), 0.5, 0.5)
        >>>         for pitch_name in 'c f d g'.split(' ')
        >>>    ]
        >>> )
        >>> for consonants, vowel, note in zip([[], [], ['t'], []], ['a', 'o', 'e', 'a'], notes):
        >>>     note.vowel = vowel
        >>>     note.consonants = consonants
        >>> isis_score_converter = isis.IsisScoreConverter('my_singing_score')
        >>> isis_score_converter.convert(notes)
        """

        extracted_data_per_event = self._convert_event(event_to_convert, 0)
        lyrics_section = self._make_lyrics_section_from_extracted_data_per_event(
            extracted_data_per_event  # type: ignore
        )
        score_section = self._make_score_section_from_extracted_data_per_event(
            extracted_data_per_event  # type: ignore
        )
        with open(self.path, "w") as f:
            f.write("\n\n".join([lyrics_section, score_section]))


class IsisConverter(converters.abc.Converter):
    """Generate audio files with `ISiS <https://forum.ircam.fr/projects/detail/isis/>`_.

    :param path: where to write the sound file
    :param isis_score_converter: The :class:`IsisScoreConverter` that shall be used
        to render the ISiS score file from a mutwo event.
    :param *flag: Flag that shall be added when calling ISiS. Several of the supported
        ISiS flags can be found in :mod:`mutwo.converters.frontends.isis_constants`.
    :param remove_score_file: Set to True if :class:`IsisConverter` shall remove the
        ISiS score file after rendering. Defaults to False.

    **Disclaimer:** Before using the :class:`IsisConverter`, make sure ISiS has been
    correctly installed on your system.
    """

    def __init__(
        self,
        path: str,
        isis_score_converter: IsisScoreConverter,
        *flag: str,
        remove_score_file: bool = False
    ):
        self.flags = flag
        self.path = path
        self.isis_score_converter = isis_score_converter
        self.remove_score_file = remove_score_file

    def convert(self, event_to_convert: ConvertableEvents) -> None:
        """Render sound file via ISiS from mutwo event.

        :param event_to_convert: The event that shall be rendered.


        **Disclaimer:** Before using the :class:`IsisConverter`, make sure
        `ISiS <https://forum.ircam.fr/projects/detail/isis/>`_ has been
        correctly installed on your system.
        """

        self.isis_score_converter.convert(event_to_convert)
        command = "{} -m {} -o {}".format(
            isis_constants.ISIS_PATH, self.isis_score_converter.path, self.path,
        )
        for flag in self.flags:
            command += " {} ".format(flag)

        os.system(command)

        if self.remove_score_file:
            os.remove(self.isis_score_converter.path)



"""Constants to be used for and with :mod:`mutwo.converters.frontends.ekmelily`.

Some constants are also used as default values of
:class:`~mutwo.converters.frontends.abjad.MutwoPitchToHEJIAbjadPitchConverter`
__init__ method.
"""

try:
    import quicktions as fractions  # type: ignore
except ImportError:
    import fractions  # type: ignore


from mutwo.parameters import pitches
from mutwo.parameters import pitches_constants

DEFAULT_GLOBAL_SCALE = (
    fractions.Fraction(0),
    fractions.Fraction(1),
    fractions.Fraction(2),
    fractions.Fraction(5, 2),
    fractions.Fraction(7, 2),
    fractions.Fraction(9, 2),
    fractions.Fraction(11, 2),
)
"""Default value for
:class:`~mutwo.converters.frontends.ekmelily.EkmelilyTuningFileConverter`
argument `global_scale`."""


# TODO(find glyph names in 'EkmelosGlyphNames.nam' file, provided
#      by Ekmelos 3.5 (instead of hard coding hexacodes))
PRIME_AND_EXPONENT_AND_TRADITIONAL_ACCIDENTAL_TO_ACCIDENTAL_GLYPH = {
    (
        None,
        None,
        pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
            fractions.Fraction(0, 1)
        ],
    ): (
        "#xE261"
    ),
    (
        None,
        None,
        pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
            fractions.Fraction(1, 1)
        ],
    ): (
        "#xE262"
    ),
    (
        None,
        None,
        pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
            fractions.Fraction(2, 1)
        ],
    ): (
        "#xE263"
    ),
    (
        None,
        None,
        pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
            -fractions.Fraction(1, 1)
        ],
    ): (
        "#xE260"
    ),
    (
        None,
        None,
        pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
            -fractions.Fraction(2, 1)
        ],
    ): (
        "#xE264"
    ),
    (
        5,
        1,
        pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
            fractions.Fraction(0, 1)
        ],
    ): (
        "#xE2C2"
    ),
    (
        5,
        2,
        pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
            fractions.Fraction(0, 1)
        ],
    ): (
        "#xE2C2"
    ),
    (
        5,
        3,
        pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
            fractions.Fraction(0, 1)
        ],
    ): (
        "#xE2D6"
    ),
    (
        5,
        -1,
        pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
            fractions.Fraction(0, 1)
        ],
    ): (
        "#xE2C7"
    ),
    (
        5,
        -2,
        pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
            fractions.Fraction(0, 1)
        ],
    ): (
        "#xE2D1"
    ),
    (
        5,
        -3,
        pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
            fractions.Fraction(0, 1)
        ],
    ): (
        "#xE2DB"
    ),
    (
        5,
        1,
        pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
            fractions.Fraction(1, 1)
        ],
    ): (
        "#xE2C3"
    ),
    (
        5,
        2,
        pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
            fractions.Fraction(1, 1)
        ],
    ): (
        "#xE2CD"
    ),
    (
        5,
        3,
        pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
            fractions.Fraction(1, 1)
        ],
    ): (
        "#xE2D7"
    ),
    (
        5,
        -1,
        pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
            fractions.Fraction(1, 1)
        ],
    ): (
        "#xE2C8"
    ),
    (
        5,
        -2,
        pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
            fractions.Fraction(1, 1)
        ],
    ): (
        "#xE2D2"
    ),
    (
        5,
        -3,
        pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
            fractions.Fraction(1, 1)
        ],
    ): (
        "#xE2DC"
    ),
    (
        5,
        1,
        pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
            fractions.Fraction(2, 1)
        ],
    ): (
        "#xE2C4"
    ),
    (
        5,
        2,
        pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
            fractions.Fraction(2, 1)
        ],
    ): (
        "#xE2CE"
    ),
    (
        5,
        3,
        pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
            fractions.Fraction(2, 1)
        ],
    ): (
        "#xE2D8"
    ),
    (
        5,
        -1,
        pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
            fractions.Fraction(2, 1)
        ],
    ): (
        "#xE2C9"
    ),
    (
        5,
        -2,
        pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
            fractions.Fraction(2, 1)
        ],
    ): (
        "#xE2D3"
    ),
    (
        5,
        -3,
        pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
            fractions.Fraction(2, 1)
        ],
    ): (
        "#xE2DD"
    ),
    (
        5,
        1,
        pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
            -fractions.Fraction(1, 1)
        ],
    ): (
        "#xE2C1"
    ),
    (
        5,
        2,
        pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
            -fractions.Fraction(1, 1)
        ],
    ): (
        "#xE2CB"
    ),
    (
        5,
        3,
        pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
            -fractions.Fraction(1, 1)
        ],
    ): (
        "#xE2D5"
    ),
    (
        5,
        -1,
        pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
            -fractions.Fraction(1, 1)
        ],
    ): (
        "#xE2C6"
    ),
    (
        5,
        -2,
        pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
            -fractions.Fraction(1, 1)
        ],
    ): (
        "#xE2D0"
    ),
    (
        5,
        -3,
        pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
            -fractions.Fraction(1, 1)
        ],
    ): (
        "#xE2DA"
    ),
    (
        5,
        1,
        pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
            -fractions.Fraction(2, 1)
        ],
    ): (
        "#xE2C0"
    ),
    (
        5,
        2,
        pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
            -fractions.Fraction(2, 1)
        ],
    ): (
        "#xE2CA"
    ),
    (
        5,
        3,
        pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
            -fractions.Fraction(2, 1)
        ],
    ): (
        "#xE2D4"
    ),
    (
        5,
        -1,
        pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
            -fractions.Fraction(2, 1)
        ],
    ): (
        "#xE2C5"
    ),
    (
        5,
        -2,
        pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
            -fractions.Fraction(2, 1)
        ],
    ): (
        "#xE2CF"
    ),
    (
        5,
        -3,
        pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
            -fractions.Fraction(2, 1)
        ],
    ): (
        "#xE2D9"
    ),
    (7, 1, None): "#xE2DE",
    (7, 2, None): "#xE2E0",
    (7, -1, None): "#xE2DF",
    (7, -2, None): "#xE2E1",
    (11, 1, None): "#xE2E3",
    (11, -1, None): "#xE2E2",
    (13, 1, None): "#xE2E4",
    (13, -1, None): "#xE2E5",
    (17, 1, None): "#xE2E6",
    (17, -1, None): "#xE2E7",
    (19, 1, None): "#xE2E9",
    (19, -1, None): "#xE2E8",
    (23, 1, None): "#xE2EA",
    (23, -1, None): "#xE2EB",
}
"""Mapping of prime, exponent and pythagorean accidental to accidental
glyph name in Ekmelos."""


TEMPERED_ACCIDENTAL_TO_ACCIDENTAL_GLYPH = {
    pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
        fractions.Fraction(0, 1)
    ]: (
        "#xE2F2"
    ),
    pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
        fractions.Fraction(1, 1)
    ]: (
        "#xE2F3"
    ),
    pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
        fractions.Fraction(2, 1)
    ]: (
        "#xE2F4"
    ),
    pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
        -fractions.Fraction(1, 1)
    ]: (
        "#xE2F1"
    ),
    pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
        -fractions.Fraction(2, 1)
    ]: (
        "#xE2F0"
    ),
    pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
        fractions.Fraction(1, 2)
    ]: (
        "#xE2F6"
    ),
    pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
        -fractions.Fraction(1, 2)
    ]: (
        "#xE2F5"
    ),
}
"""Mapping of tempered accidental name to glyph name in Ekmelos."""

TEMPERED_ACCIDENTAL_TO_CENT_DEVIATION = {
    pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
        fractions.Fraction(0, 1)
    ]: 0,
    pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
        fractions.Fraction(1, 1)
    ]: 100,
    pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
        fractions.Fraction(2, 1)
    ]: 200,
    pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
        -fractions.Fraction(1, 1)
    ]: -100,
    pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
        -fractions.Fraction(2, 1)
    ]: -200,
    pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
        fractions.Fraction(1, 2)
    ]: 50,
    pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
        -fractions.Fraction(1, 2)
    ]: -50,
}
"""Mapping of tempered accidental name to cent deviation."""

DEFAULT_PRIME_TO_HEJI_ACCIDENTAL_NAME = {
    prime: accidental_name
    for prime, accidental_name in zip(
        (5, 7, 11, 13, 17, 19, 23), "a b c d e f g".split(" ")
    )
}
"""Default mapping for
:class:`~mutwo.converters.frontends.ekmelily.HEJIEkmelilyTuningFileConverter`
argument `prime_to_heji_accidental_name`."""

DEFAULT_PRIME_TO_HIGHEST_ALLOWED_EXPONENT = {
    prime: highest_allowed_comma
    # all potentially supported prime / max_exponent pairs:
    # (not used by default, because Lilypond would take too
    #  long for parsing)
    # for prime, highest_allowed_comma in zip(
    #     (5, 7, 11, 13, 17, 19, 23), (3, 2, 1, 1, 1, 1, 1),
    # )
    for prime, highest_allowed_comma in zip((5, 7, 11, 13, 17), (3, 2, 1, 1, 1),)
}
"""Default value for
:class:`~mutwo.converters.frontends.ekmelily.HEJIEkmelilyTuningFileConverter`
argument `prime_to_highest_allowed_exponent`."""

DEFAULT_TEMPERED_PITCH_INDICATOR = "t"
"""Default value for
:class:`~mutwo.converters.frontends.ekmelily.HEJIEkmelilyTuningFileConverter`
argument `tempered_pitch_indicator`."""

DEFAULT_OTONALITY_INDICATOR = "o"
"""Default value for
:class:`~mutwo.converters.frontends.ekmelily.HEJIEkmelilyTuningFileConverter`
argument `otonality_indicator`."""

DEFAULT_UTONALITY_INDICATOR = "u"
"""Default value for
:class:`~mutwo.converters.frontends.ekmelily.HEJIEkmelilyTuningFileConverter`
argument `utonality_indicator`."""

# solution from: https://stackoverflow.com/questions/23199733/convert-numbers-into-corresponding-letter-using-python
DEFAULT_EXPONENT_TO_EXPONENT_INDICATOR = lambda exponent: chr(ord("a") + exponent)
"""Default function for
:class:`~mutwo.converters.frontends.ekmelily.HEJIEkmelilyTuningFileConverter`
argument `exponent_to_exponent_indicator`."""

PYTHAGOREAN_ACCIDENTAL_CENT_DEVIATION_SIZE = round(
    (pitches.JustIntonationPitch((0, 7)).normalize(mutate=False).cents), 2  # type: ignore
)
"""Step in cents for one pythagorean accidental (# or b)."""

PYTHAGOREAN_ACCIDENTAL_TO_CENT_DEVIATIONS = {
    pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
        fractions.Fraction(0, 1)
    ]: 0,
    pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
        fractions.Fraction(1, 1)
    ]: PYTHAGOREAN_ACCIDENTAL_CENT_DEVIATION_SIZE,
    pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
        fractions.Fraction(2, 1)
    ]: PYTHAGOREAN_ACCIDENTAL_CENT_DEVIATION_SIZE
    * 2,
    pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
        -fractions.Fraction(1, 1)
    ]: -PYTHAGOREAN_ACCIDENTAL_CENT_DEVIATION_SIZE,
    pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME[
        -fractions.Fraction(2, 1)
    ]: -2
    * PYTHAGOREAN_ACCIDENTAL_CENT_DEVIATION_SIZE,
}
"""Step in cents mapping for each pythagorean accidental (# or b)."""

DIFFERENCE_BETWEEN_PYTHAGOREAN_AND_TEMPERED_FIFTH = (
    pitches.JustIntonationPitch("3/2").cents - 700
)
"""The difference in cents between a just fifth (3/2) and
a 12-EDO fifth. This constant is used in
:class:`~mutwo.converters.frontends.ekmelily.HEJIEkmelilyTuningFileConverter`.
"""



"""Time-based Event abstractions.

Event objects can be understood as the core objects
of the :mod:`mutwo` framework. They all own a :attr:`duration`
attribute (which can be any number). Further more complex
Event classes with more relevant attributes can be generated
through inheriting from basic classes. :mod:`mutwo` already offers
support for several more complex representations (for instance
:class:`mutwo.events.music.NoteLike`).

The most often used classes may be:
    - :class:`mutwo.events.basic.SimpleEvent`
    - :class:`mutwo.events.basic.SequentialEvent`
    - :class:`mutwo.events.basic.SimultaneousEvent`
"""

from . import music_constants

from . import abc
from . import basic
from . import music
from . import general



"""Algorithms which are related to Scottish botanist Robert Brown."""

import math
import typing

import numpy as np  # type: ignore
from scipy.stats import norm  # type: ignore

__all__ = ("random_walk_noise",)


def random_walk_noise(
    x0: float,
    n: int,
    dt: float,
    delta: float,
    out: typing.Union[np.array, None] = None,
    random_state: int = None,
) -> np.array:
    """Generate an instance of Brownian motion (i.e. the Wiener process).

    :param x0: the initial condition(s) (i.e. position(s)) of the Brownian motion.
    :param n: the number of steps to take
    :param dt: the time step
    :param delta: delta determines the "speed" of the Brownian motion.  The random variable
        of the position at time t, X(t), has a normal distribution whose mean is
        the position at time t=0 and whose variance is delta**2*t.
    :param out: If `out` is not None, it specifies the array in which to put the
        result.  If `out` is None, a new numpy array is created and returned.
    :param random_state: set the random seed of the pseudo-random generator.

    :return: A numpy array of floats with shape `x0.shape + (n,)`.

    X(t) = X(0) + N(0, delta**2 * t; 0, t)

    where N(a,b; t0, t1) is a normally distributed random variable with mean a and
    variance b.  The parameters t0 and t1 make explicit the statistical
    independence of N on different time intervals; that is, if [t0, t1) and
    [t2, t3) are disjoint intervals, then N(a, b; t0, t1) and N(a, b; t2, t3)
    are independent.

    Written as an iteration scheme,

        X(t + dt) = X(t) + N(0, delta**2 * dt; t, t+dt)

    If `x0` is an array (or array-like), each value in `x0` is treated as
    an initial condition, and the value returned is a numpy array with one
    more dimension than `x0`.

    Note that the initial value `x0` is not included in the returned array.

    This code has been copied from the scipy cookbook:
        https://scipy-cookbook.readthedocs.io/items/BrownianMotion.html
    """

    x0_as_array = np.asarray(x0)

    # For each element of x0, generate a sample of n numbers from a
    # normal distribution.
    r = norm.rvs(
        size=x0_as_array.shape + (n,), scale=delta * math.sqrt(dt), random_state=random_state
    )

    # If `out` was not given, create an output array.
    if out is None:
        out = np.empty(r.shape)

    # This computes the Brownian motion by forming the cumulative sum of
    # the random samples.
    np.cumsum(r, axis=-1, out=out)

    # Add the initial condition.
    out += np.expand_dims(x0_as_array, axis=-1)

    return out



"""Abstract base classes for events (definition of public API)."""

from __future__ import annotations

import abc
import copy
import typing

from mutwo import parameters
from mutwo.utilities import decorators
from mutwo.utilities import tools

__all__ = ("Event", "ComplexEvent")


class Event(abc.ABC):
    """Abstract Event-Object"""

    # ###################################################################### #
    #                           properties                                   #
    # ###################################################################### #

    @property
    @abc.abstractmethod
    def duration(self) -> parameters.abc.DurationType:
        """The duration of an event (which can be any number).

        The unit of the duration is up to the interpretation of the user
        and the respective conversion routine that will be used. For
        instance when using :class:`CsoundScoreConverter`, the duration will be
        understood as seconds, while :class:`MidiFileConverter` will read duration
        as beats.
        """
        raise NotImplementedError

    # ###################################################################### #
    #                           private methods                              #
    # ###################################################################### #

    @staticmethod
    def _assert_correct_start_and_end_values(
        start: parameters.abc.DurationType,
        end: parameters.abc.DurationType,
        condition: typing.Callable[
            [parameters.abc.DurationType, parameters.abc.DurationType], bool
        ] = lambda start, end: end
        >= start,
    ):
        """Helper method to make sure that start < end.

        Can be used within the different cut_out methods.
        """
        try:
            assert condition(start, end)
        except AssertionError:
            message = (
                "Invalid values for start and end property (start = '{}' and end ="
                " '{}')!".format(start, end)
            )
            message += " Value for end has to be bigger than value for start."
            raise ValueError(message)

    # ###################################################################### #
    #                           public methods                               #
    # ###################################################################### #

    @abc.abstractmethod
    def destructive_copy(self) -> Event:
        """Adapted deep copy method that returns a new object for every leaf.

        It's called 'destructive', because it forgets potential repetitions of
        the same object in compound objects. Instead of reproducing the original
        structure of the compound object that shall be copied, every repetition
        of the same reference will return a new unique independent object.

        The following example shall illustrate the difference between copy.deepcopy
        and destructive_copy:

        >>> import copy
        >>> from mutwo.events import basic
        >>> my_simple_event_0 = basic.SimpleEvent(2)
        >>> my_simple_event_1 = basic.SimpleEvent(3)
        >>> my_sequential_event = basic.SequentialEvent([my_simple_event_0, my_simple_event_1, my_simple_event_0])
        >>> deepcopied_event = copy.deepcopy(my_sequential_event)
        >>> destructivecopied_event = my_sequential_event.destructive_copy()
        >>> deepcopied_event[0].duration = 10  # setting the duration of the first event
        >>> destructivecopied_event[0].duration = 10
        >>> # return True because the first and the third objects share the same
        >>> # reference (both are the same copy of 'my_simple_event_0')
        >>> deepcopied_event[0].duration == deepcopied_event[2].duration
        True
        >>> # return False because destructive_copy forgets the shared reference
        >>> destructivecopied_event[0].duration == destructivecopied_event[2].duration
        False
        """

        raise NotImplementedError

    @abc.abstractmethod
    def get_parameter(
        self, parameter_name: str
    ) -> typing.Union[typing.Tuple[typing.Any, ...], typing.Any]:
        """Return event attribute with the entered name.

        :param parameter_name: The name of the attribute that shall be returned.
        :returns: Return tuple containing the assigned values for each contained
            event. If an event doesn't posses the asked parameter, mutwo will simply
            add None to the tuple for the respective event.

        **Example:**

        >>> from mutwo.events import basic
        >>> sequential_event = basic.SequentialEvent([basic.SimpleEvent(2), basic.SimpleEvent(3)])
        >>> sequential_event.get_parameter('duration')
        (2, 3)
        """

        raise NotImplementedError

    @abc.abstractmethod
    def set_parameter(
        self,
        parameter_name: str,
        object_or_function: typing.Union[
            typing.Callable[
                [parameters.abc.ParameterType], parameters.abc.ParameterType
            ],
            typing.Any,
        ],
        set_unassigned_parameter: bool = True,
    ) -> None:
        """Sets parameter to new value for all children events.

        :param parameter_name: The name of the parameter which values shall be changed.
        :param object_or_function: For setting the parameter either a new value can be
            passed directly or a function can be passed. The function gets as an
            argument the previous value that has had been assigned to the respective
            object and has to return a new value that will be assigned to the object.
        :param set_unassigned_parameter: If set to False a new parameter will only be
            assigned to an Event if the Event already has a attribute with the
            respective `parameter_name`. If the Event doesn't know the attribute yet
            and `set_unassigned_parameter` is False, the method call will simply be
            ignored.

        **Example:**

        >>> from mutwo.events import basic
        >>> sequential_event = basic.SequentialEvent([basic.SimpleEvent(2), basic.SimpleEvent(3)])
        >>> sequential_event.set_parameter('duration', lambda duration: duration * 2)
        >>> sequential_event.get_parameter('duration')
        (4, 6)
        """

        raise NotImplementedError

    @abc.abstractmethod
    def mutate_parameter(
        self,
        parameter_name: str,
        function: typing.Union[
            typing.Callable[[parameters.abc.Parameter], None], typing.Any
        ],
    ) -> None:
        """Mutate parameter with a function.

        :param parameter_name: The name of the parameter which shall be mutated.
        :param function: The function which mutates the parameter. The function gets
            as an input the assigned value for the passed parameter_name of the
            respective object. The function shouldn't return anything, but simply
            calls a method of the parameter value.

        This method is useful when a particular parameter has been assigned to objects
        that know methods which mutate themselves. Then 'mutate_parameter' is a
        convenient wrapper to call the methods of those parameters for all children
        events.

        **Example:**

        >>> from mutwo.events import basic, music
        >>> from mutwo.parameters import pitches
        >>> sequential_event = basic.SequentialEvent([music.NoteLike([pitches.WesternPitch('c', 4), pitches.WesternPitch('e', 4)], 2, 1)])
        >>> sequential_event.mutate_parameter('pitch_or_pitches', lambda pitch_or_pitches: [pitch.add(12) for pitch in pitch_or_pitches])
        >>> # now all pitches should be one octave higher (from 4 to 5)
        >>> sequential_event.get_parameter('pitch_or_pitches')
        ([WesternPitch(c5), WesternPitch(e5)],)
        """

        raise NotImplementedError

    @abc.abstractmethod
    def cut_out(
        self, start: parameters.abc.DurationType, end: parameters.abc.DurationType,
    ) -> typing.Optional[Event]:
        """Time-based slicing of the respective event.

        :param start: number that indicates the point when the
            cut out shall start.
        :param end: number that indicates the point when the cut
            up shall end.

        **Example:**

        >>> from mutwo.events import basic
        >>> sequential_event = basic.SequentialEvent([basic.SimpleEvent(3), basic.SimpleEvent(2)])
        >>> sequential_event.cut_out(1, 4)
        >>> print(sequential_event)
        SequentialEvent([SimpleEvent(duration = 2), SimpleEvent(duration = 1)])
        """

        raise NotImplementedError

    @abc.abstractmethod
    def cut_off(
        self, start: parameters.abc.DurationType, end: parameters.abc.DurationType,
    ) -> typing.Optional[Event]:
        """Time-based deletion / shortening of the respective event.

        :param start: number that indicates absolute time when the
            cut off shall start.
        :param end: number that indicates the absolute time when the cut
            off shall end.

        **Example:**

        >>> from mutwo.events import basic
        >>> sequential_event = basic.SequentialEvent([basic.SimpleEvent(3), basic.SimpleEvent(2)])
        >>> sequential_event.cut_off(1, 3)
        >>> print(sequential_event)
        SequentialEvent([SimpleEvent(duration = 1), SimpleEvent(duration = 1)])
        """

        raise NotImplementedError

    def split_at(
        self, absolute_time: parameters.abc.DurationType
    ) -> typing.Tuple[Event, Event]:
        """Split event in two events at :attr:`absolute_time`.

        :param absolute_time: where event shall be split
        :return: Two events that result from splitting the present event.

        **Example:**

        >>> from mutwo.events import basic
        >>> sequential_event = basic.SequentialEvent([basic.SimpleEvent(3)])
        >>> sequential_event.split_at(1)
        (SequentialEvent([SimpleEvent(duration = 1)]), SequentialEvent([SimpleEvent(duration = 2)]))
        >>> sequential_event[0].split_at(1)
        (SimpleEvent(duration = 1), SimpleEvent(duration = 2))
        """

        return (
            self.cut_out(0, absolute_time, mutate=False),  # type: ignore
            self.cut_out(absolute_time, self.duration, mutate=False),  # type: ignore
        )


T = typing.TypeVar("T", bound=Event)


class ComplexEvent(Event, typing.List[T], typing.Generic[T]):
    """Abstract Event-Object, which contains other Event-Objects."""

    def __init__(self, iterable: typing.Iterable[T]):
        super().__init__(iterable)

    # ###################################################################### #
    #                           magic methods                                #
    # ###################################################################### #

    def __repr__(self) -> str:
        return "{}({})".format(type(self).__name__, super().__repr__())

    def __add__(self, event: typing.List[T]) -> ComplexEvent[T]:
        return type(self)(super().__add__(event))

    def __mul__(self, factor: int) -> ComplexEvent[T]:
        return type(self)(super().__mul__(factor))

    @typing.overload
    def __getitem__(self, index_or_slice: int) -> T:
        ...

    @typing.overload
    def __getitem__(self, index_or_slice: slice) -> ComplexEvent[T]:
        ...

    def __getitem__(
        self, index_or_slice: typing.Union[int, slice]
    ) -> typing.Union[T, ComplexEvent[T]]:
        event = super().__getitem__(index_or_slice)
        if isinstance(index_or_slice, slice):
            event = type(self)(event)  # type: ignore
        return event  # type: ignore

    # ###################################################################### #
    #                           properties                                   #
    # ###################################################################### #

    @Event.duration.setter  # type: ignore
    def duration(self, new_duration: parameters.abc.DurationType):
        old_duration = self.duration
        self.set_parameter(
            "duration",
            lambda duration: tools.scale(duration, 0, old_duration, 0, new_duration),  # type: ignore
        )

    # ###################################################################### #
    #                           private methods                              #
    # ###################################################################### #

    def _assert_start_in_range(self, start: parameters.abc.DurationType):
        """Helper method to make sure that start < event.duration.

        Can be used within the different squash_in methods.
        """
        try:
            assert self.duration >= start
        except AssertionError:
            message = (
                "Invalid value for start = '{}' in 'squash_in' call for event with"
                " duration = '{}'!".format(start, self.duration)
            )
            message += " Start has to be equal or smaller than the events duration."
            raise ValueError(message)

    # ###################################################################### #
    #                           public methods                               #
    # ###################################################################### #

    def copy(self) -> ComplexEvent[T]:
        """Return a deep copy of the ComplexEvent."""
        return copy.deepcopy(self)

    def destructive_copy(self) -> ComplexEvent[T]:
        return type(self)([event.destructive_copy() for event in self])  # type: ignore

    def get_event_from_indices(self, indices: typing.Sequence[int]) -> Event:
        """Get nested :class:`Event` from a sequence of indices.

        :param indices: The indices of the nested :class:`Event`.
        :type indices: typing.Sequence[int]

        **Example:**

        >>> from mutwo.events import basic
        >>> nested_sequential_event = basic.SequentialEvent([basic.SequentialEvent([basic.SimpleEvent(2)])])
        >>> nested_sequential_event.get_event_from_indices((0, 0))
        SimpleEvent(duration = 2)
        >>> # this is equal to:
        >>> nested_sequential_event[0][0]
        SimpleEvent(duration = 2)
        """

        return tools.get_nested_item_from_indices(indices, self)

    def get_parameter(self, parameter_name: str) -> typing.Tuple[typing.Any, ...]:
        return tuple(event.get_parameter(parameter_name) for event in self)

    @decorators.add_return_option
    def set_parameter(  # type: ignore
        self,
        parameter_name: str,
        object_or_function: typing.Union[
            typing.Callable[
                [parameters.abc.ParameterType], parameters.abc.ParameterType
            ],
            typing.Any,
        ],
        set_unassigned_parameter: bool = True,
    ) -> typing.Optional[ComplexEvent[T]]:
        [
            event.set_parameter(
                parameter_name,
                object_or_function,
                set_unassigned_parameter=set_unassigned_parameter,
            )
            for event in self
        ]

    @decorators.add_return_option
    def mutate_parameter(  # type: ignore
        self,
        parameter_name: str,
        function: typing.Union[
            typing.Callable[[parameters.abc.Parameter], None], typing.Any
        ],
    ) -> typing.Optional[ComplexEvent[T]]:
        [event.mutate_parameter(parameter_name, function) for event in self]

    @abc.abstractmethod
    def squash_in(
        self, start: parameters.abc.DurationType, event_to_squash_in: Event
    ) -> typing.Optional[ComplexEvent[T]]:
        """Time-based insert of a new event into the present event.

        :param start: Absolute time where the event shall be inserted.
        :param event_to_squash_in: the event that shall be squashed into
            the present event.

        Squash in a new event to the present event.

        **Example:**

        >>> from mutwo.events import basic
        >>> sequential_event = basic.SequentialEvent([basic.SimpleEvent(3)])
        >>> sequential_event.squash_in(1, basic.SimpleEvent(1.5))
        >>> print(sequential_event)
        SequentialEvent([SimpleEvent(duration = 1), SimpleEvent(duration = 1.5), SimpleEvent(duration = 0.5)])
        """

        raise NotImplementedError



"""Algorithms which are related to Canadian computer scientist G. T. Toussaint."""

import itertools
import typing

from mutwo.utilities import tools

__all__ = ("euclidean", "paradiddle", "alternating_hands")


def euclidean(size: int, distribution: int) -> typing.Tuple[int, ...]:
    """Return euclidean rhythm as described in a 2005 paper by G. T. Toussaint.

    :param size: how many beats the rhythm contains
    :param distribution: how many beats are played
    :return: The rhythm in relative time.

    **Example:**

    >>> from mutwo.generators import toussaint
    >>> toussaint.euclidean(8, 4)
    (2, 2, 2, 2)
    >>> toussaint.euclidean(7, 5)
    (2, 1, 1, 2, 1)

    The title of Toussaints paper is
    "The Euclidean Algorithm Generates Traditional Musical Rhythms".
    """

    standard_size = size // distribution
    rest = size % distribution
    data = (standard_size for _ in range(distribution))
    if rest:
        added = tuple(tools.accumulate_from_zero(euclidean(distribution, rest)))
        return tuple(s + 1 if idx in added else s for idx, s in enumerate(data))
    else:
        return tuple(data)


def _mirror(pattern: typing.Tuple[bool, ...]) -> typing.Tuple[bool, ...]:
    """Inverse every boolean value inside the tuple.

    Helper function for other functions.
    """
    return tuple(False if item else True for item in pattern)


def paradiddle(size: int) -> typing.Tuple[typing.Tuple[int, ...], ...]:
    """Generates rhythm using the paradiddle method described by G. T. Toussaint.

    :param size: how many beats the resulting rhythm shall last. 'Size' has to be
        divisible by 2 because of the symmetrical structure of the generated rhythm.
    :return: Return nested tuple that contains two tuple where each tuple represents
        one rhythm (both rhythms are complementary to each other). The rhythms are
        encoded in absolute time values.

    **Example:**

    >>> from mutwo.generators import toussaint
    >>> toussaint.paradiddle(8)
    ((0, 2, 3, 5), (1, 4, 6, 7))
    >>> toussaint.paradiddle(6)
    ((0, 4, 5), (1, 2, 3))

    The paradiddle algorithm has been described by Godfried T. Toussaint
    in his paper 'Generating “Good” Musical Rhythms Algorithmically'.
    """

    def convert_to_right_left_pattern(pattern: tuple) -> tuple:
        right = []
        left = []
        for idx, item in enumerate(pattern):
            if item:
                right.append(idx)
            else:
                left.append(idx)
        return tuple(right), tuple(left)

    # check for correct size value
    try:
        assert size % 2 == 0 and size > 2
    except AssertionError:
        message = (
            "Invalid value '{}' for argument 'size'. 'Size' has to be divisible by 2"
            " and has to be bigger than 2.".format(size)
        )
        raise ValueError(message)

    cycle = itertools.cycle((True, False))
    pattern = list(next(cycle) for _ in range(size // 2))
    pattern[-1] = pattern[-2]
    return convert_to_right_left_pattern(tuple(pattern) + _mirror(tuple(pattern)))


def alternating_hands(
    seed_rhythm: typing.Tuple[int, ...],
) -> typing.Tuple[typing.Tuple[int, ...], ...]:
    """Generates rhythm using the alternating hands method described by G. T. Toussaint.

    :param seed_rhythm: rhythm that shall be distributed on two hands.
    :return: Return nested tuple that contains two tuple where each tuple represents
        one rhythm (both rhythms are complementary to each other). The rhythms are
        encoded in absolute time values.

    **Example:**

    >>> from mutwo.generators import toussaint
    >>> toussaint.alternating_hands((2, 2))
    ((0, 6), (2, 4))
    >>> toussaint.alternating_hands((3, 2, 2))
    ((0, 5, 10), (3, 7, 12))

    The alternating hands algorithm has been described by Godfried T. Toussaint
    in his paper 'Generating “Good” Musical Rhythms Algorithmically'.
    """

    n_elements = len(seed_rhythm)
    absolute_rhythm = tuple(tools.accumulate_from_zero(seed_rhythm + seed_rhythm))
    cycle = itertools.cycle((True, False))
    distribution = tuple(next(cycle) for n in range(n_elements))
    distribution += _mirror(distribution)
    right, left = [], []
    for idx, dis in enumerate(distribution):
        item = absolute_rhythm[idx]
        if dis:
            right.append(item)
        else:
            left.append(item)
    return tuple(right), tuple(left)



"""Convert mutwo objects to external data."""

from mutwo.utilities import tools

from . import csound_constants
from . import csound
from . import ekmelily_constants
from . import ekmelily
from . import isis_constants
from . import isis

# import modules with extra require
for modules, dependencies in (
    (("midi_constants", "midi"), ("mido",)),
    (("reaper",), ("rpp",)),
    (("abjad_attachments", "abjad_constants", "abjad",), ("abjad", "abjadext.nauert")),
):
    for module in modules:
        tools.import_module_if_dependencies_have_been_installed(
            "mutwo.converters.frontends.{}".format(module), dependencies
        )

del module, dependencies, tools



import typing


class VstInfo(object):
    def __init__(
        self,
        filename: str,
        binary: typing.Sequence[str],
        parameters: typing.Mapping[str, str],
    ):
        self.filename = filename
        self.binary = binary
        self.parameters = parameters


GRAILLON2 = VstInfo(
    "Auburn Sounds Graillon 2.vst3",
    [
        "qnXPfe5e7f4CAAAAAQAAAAAAAAACAAAAAAAAAAIAAAABAAAAAAAAAAIAAAAAAAAAwgAAAAEAAAAAABAA",
        "sgAAAAEAAAAAAAsgupIAAAAAAAAAAAAFAgAAAAAAJgAAAO87NT8AAAAAAAAAAAAAAAAAAAAAAACAPwAAAAAAAAAA7zs1PwAAAAAAAAAAAAAAAAAAgD8AAAA/AAAAAAAA",
        "AD8AAAAAAAAAAAAAAAAAAIA/0zIBPwAAAAAAAIA/AACAPwAAgD8AAIA/AACAPwAAgD8AAIA/AACAPwAAgD8AAIA/AACAPwAAgD8AAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
        "AAA=",
        "AEZhY3RvcnkgUHJlc2V0czogRGVmYXVsdAAQAAAA",
    ],
    {
        "bitdepth": "12:10",
        "quantize": "13:11",
        "pitch shift": "17:15",
        "preserve formants": "18:16",
        "correction": "19:17",
        "pitch smoothing": "20:18",
        # "allow c": "24:22",
    },
)

COLOURCOPY = VstInfo(
    "ColourCopy.vst3",
    [
        "nn/9SO5e7f4CAAAAAQAAAAAAAAACAAAAAAAAAAIAAAABAAAAAAAAAAIAAAAAAAAAyAkAAAEAAAAAABAA",
        "3AQAAAEAAADYBAAAI3BnbT1pbml0aWFsaXplLmgycAojQU09Q29sb3VyQ29weQojVmVycz0xMDAwMAojRW5kaWFuPWxpdHRsZQojbm09OAojbXM9bm9uZQojbXM9TW9k",
        "V2hsCiNtcz1QaXRjaFcKI21zPUN0cmxBCiNtcz1DdHJsQgojbXM9RW52CiNtcz1MRk8gTAojbXM9TEZPIFIKI252PTAKI2NtPVBDb3JlClNCYXNlPTQKU3dpbmc9MC4w",
        "MApTVHJpZz0xClBQcm9qPTAKUEZvbGQ9MApQRmlsZT0xCkdGaWxlPTIKR1NjYWxlPTAKQ2hMYXk9MApTdXJyTz0xClJldj05Nzc1CkxFRD0wLjAwClBBR0U9MApQYWdl",
        "c09uPTAKQ29yZU49MwpTbGljZT00ClVJX29wPTUKTWlkaUE9NQpNaWRpUD02CkRGb2xkPTcKQ3RybEE9MgpDdHJsQj0xMQojY209TEZPClJhdGU9NTAuMDAKU3RlcmVv",
        "PTkwLjAwCkZNb2Q9MC4wMApXYXZlPTAKI2NtPURlbGF5ClRpbWVCPTgKRGx5UmF0ZT0wLjAwCk5vdGVDdHI9MApUYXBMPTEwMC4wMApUYXBSPTEwMC4wMApMRk9Ucmc9",
        "MApMRk9EcHQ9MC4wMApMRk9EeW49MC4wMApSZWdlbj01MC4wMApJbnZlcnQ9MApCcmlnaHQ9MC4wMApTYXQ9NTAuMDAKQ29sb3VyPTMuMDAKRmxvb3I9MQpEdWNrPTAK",
        "RHVja1RyZz0wCkR1Y2tUaHI9MC4wMApSb3V0SU49MApSb3V0RkI9MApXaWR0aD0xMDAuMDAKTWl4PTUwLjAwCkluRz0wLjAwCk91dEc9MC4wMApQYW5pYz0wCkZyZWV6",
        "ZT0wCkJ5cGFzcz0wCkhRPTAKI2NtPUxNZW0KTGF5TWVtMT0wCiNjbT1TdGlja2VyCnN0a3JUeHQ9OApiZ2lkeD05CnBvc3g9LTEuMDAKcG9zeT0tMS4wMAoKCgoKLy8g",
        "U2VjdGlvbiBmb3IgdWdseSBjb21wcmVzc2VkIGJpbmFyeSBEYXRhCi8vIERPTidUIFRPVUNIIFRISVMKCiQkJCQxMzMzCj9nZGdjZW5lYjphYWFiYWFhYTplYWFhYWFh",
        "YTpnamdvZ2poZTpnamdiZ21najpoa2dmY29naTpkY2hhYWFhYToKb2JjanBjZGw6YW9oZGlvZmE6Y2xsamluZ2I6YWE6bW06YWU6Y2o6ZGg6aWk6aXA6a2Q6YWQ6YWg6",
        "YWk6YW46YgpiOmJnOmRiOmRrOmVhOmViOmVqOmVuOmZpOmZqOmZvOmdhOmdjOmdkIVpZVFJBYWJBMjY1WllUUlFBMTBnamdvCmdqaGVnamdiZ21namhrZ2Zjb2dpZGNo",
        "YUE0OVpZVFJRQTc0WllUUlFBMTBvYkRwY2RsYW9oZGlvZmFjbGxqaW4KZ2JYRmtla25tY0lsYXBtVVZva1dobGhibG9ESExCb2Jua2trU21hZ2tPR2hmbWhvcERCbWZH",
        "bHBpZGxoQ2xuQgpLbmhIaWZrYWhlTnBvRVBsY0ZaWVRSQ0ExMG1iRUpBWllUUkFhYkEyNjVaWVRSaWFBMTM4WllUUkFhYkEyNjVaCllUUk1BMjdaWVRSUUE3ND0zNTk1",
        "MwoAANwEAAAAAAAA2AQAACNwZ209aW5pdGlhbGl6ZS5oMnAKI0FNPUNvbG91ckNvcHkKI1ZlcnM9MTAwMDAKI0VuZGlhbj1saXR0bGUKI25tPTgKI21zPW5vbmUKI21z",
        "PU1vZFdobAojbXM9UGl0Y2hXCiNtcz1DdHJsQQojbXM9Q3RybEIKI21zPUVudgojbXM9TEZPIEwKI21zPUxGTyBSCiNudj0wCiNjbT1QQ29yZQpTQmFzZT00ClN3aW5n",
        "PTAuMDAKU1RyaWc9MQpQUHJvaj0wClBGb2xkPTAKUEZpbGU9MQpHRmlsZT0yCkdTY2FsZT0wCkNoTGF5PTAKU3Vyck89MQpSZXY9OTc3NQpMRUQ9MC4wMApQQUdFPTAK",
        "UGFnZXNPbj0wCkNvcmVOPTMKU2xpY2U9NApVSV9vcD01Ck1pZGlBPTUKTWlkaVA9NgpERm9sZD03CkN0cmxBPTIKQ3RybEI9MTEKI2NtPUxGTwpSYXRlPTUwLjAwClN0",
        "ZXJlbz05MC4wMApGTW9kPTAuMDAKV2F2ZT0wCiNjbT1EZWxheQpUaW1lQj04CkRseVJhdGU9MC4wMApOb3RlQ3RyPTAKVGFwTD0xMDAuMDAKVGFwUj0xMDAuMDAKTEZP",
        "VHJnPTAKTEZPRHB0PTAuMDAKTEZPRHluPTAuMDAKUmVnZW49NTAuMDAKSW52ZXJ0PTAKQnJpZ2h0PTAuMDAKU2F0PTUwLjAwCkNvbG91cj0zLjAwCkZsb29yPTEKRHVj",
        "az0wCkR1Y2tUcmc9MApEdWNrVGhyPTAuMDAKUm91dElOPTAKUm91dEZCPTAKV2lkdGg9MTAwLjAwCk1peD01MC4wMApJbkc9MC4wMApPdXRHPTAuMDAKUGFuaWM9MApG",
        "cmVlemU9MApCeXBhc3M9MApIUT0wCiNjbT1MTWVtCkxheU1lbTE9MAojY209U3RpY2tlcgpzdGtyVHh0PTgKYmdpZHg9OQpwb3N4PS0xLjAwCnBvc3k9LTEuMDAKCgoK",
        "Ci8vIFNlY3Rpb24gZm9yIHVnbHkgY29tcHJlc3NlZCBiaW5hcnkgRGF0YQovLyBET04nVCBUT1VDSCBUSElTCgokJCQkMTMzMwo/Z2RnY2VuZWI6YWFhYmFhYWE6ZWFh",
        "YWFhYWE6Z2pnb2dqaGU6Z2pnYmdtZ2o6aGtnZmNvZ2k6ZGNoYWFhYWE6Cm9iY2pwY2RsOmFvaGRpb2ZhOmNsbGppbmdiOmFhOm1tOmFlOmNqOmRoOmlpOmlwOmtkOmFk",
        "OmFoOmFpOmFuOmIKYjpiZzpkYjpkazplYTplYjplajplbjpmaTpmajpmbzpnYTpnYzpnZCFaWVRSQWFiQTI2NVpZVFJRQTEwZ2pnbwpnamhlZ2pnYmdtZ2poa2dmY29n",
        "aWRjaGFBNDlaWVRSUUE3NFpZVFJRQTEwb2JEcGNkbGFvaGRpb2ZhY2xsamluCmdiWEZrZWtubWNJbGFwbVVWb2tXaGxoYmxvREhMQm9ibmtra1NtYWdrT0doZm1ob3BE",
        "Qm1mR2xwaWRsaENsbkIKS25oSGlma2FoZU5wb0VQbGNGWllUUkNBMTBtYkVKQVpZVFJBYWJBMjY1WllUUmlhQTEzOFpZVFJBYWJBMjY1WgpZVFJNQTI3WllUUlFBNzQ9",
        "MzU5NTMKAAA=",
        "AHByb2dyYW06IFByb2dyYW0gMAAQAAAA",
    ],
    {
        "delay rate": "5:5",
        "tap left": "7:7",
        "tap right": "8:8",
        "regeneration": "12:12",
    },
)

TWANGSTROEM = VstInfo(
    "Twangstrom.vst3",
    [
        "dOz1BO5e7f4EAAAAAQAAAAAAAAACAAAAAAAAAAQAAAAAAAAACAAAAAAAAAACAAAAAQAAAAAAAAACAAAAAAAAAMoMAAABAAAAAAAQAA==",
        "XQYAAAEAAABZBgAAI3BnbT1pbml0aWFsaXplLmgycAojQU09VHdhbmdzdHJvbQojVmVycz0xMDAwMAojRW5kaWFuPWxpdHRsZQojbm09NwojbXM9bm9uZQojbXM9TW9k",
        "V2hsCiNtcz1QaXRjaFcKI21zPUN0cmxBCiNtcz1DdHJsQgojbXM9TEZPCiNtcz1FbnZlbG9wZQojbnY9MAojY209bWFpbgpDcmNSPTEKQ3JjUjI9MQojY209UENvcmUK",
        "U1RyaWc9MQpQUHJvaj0wClBGb2xkPTAKUEZpbGU9MQpHRmlsZT0yCkdTY2FsZT0wCkNoTGF5PTAKU3Vyck89MQpSZXY9OTc3NQpMRUQ9MC4wMApQQUdFPTAKUGFnZXNP",
        "bj0wCkNvcmVOPTMKU2xpY2U9NApVSV9vcD0zCk1pZGlBPTUKTWlkaVA9NgpERm9sZD03CkN0cmxBPTIKQ3RybEI9MTEKI2NtPU1NMQpBY3RpdmU9MQpTb3VyY2U9MApE",
        "ZXN0MT1ub25lOmFzc2lnbmVkCkRlcHRoMT0wLjAwCkN1cnZlMT0yClFudHplMT0wClJlY3QxPTAKU0gxPTAKU2xldzE9MAojY209TU0yCkFjdGl2ZT0xClNvdXJjZT0w",
        "CkRlc3QxPW5vbmU6YXNzaWduZWQKRGVwdGgxPTAuMDAKQ3VydmUxPTIKUW50emUxPTAKUmVjdDE9MApTSDE9MApTbGV3MT0wCiNjbT1NTTMKQWN0aXZlPTEKU291cmNl",
        "PTAKRGVzdDE9bm9uZTphc3NpZ25lZApEZXB0aDE9MC4wMApDdXJ2ZTE9MgpRbnR6ZTE9MApSZWN0MT0wClNIMT0wClNsZXcxPTAKI2NtPUxGTwpTeW5jPTYKV2F2ZT0x",
        "ClJhdGU9MC4wMAojY209RW52ClNvdXJjZT0wCk1vZGU9MApUaHJlcz0tMzAuMDAKQXR0YWNrPTI1LjAwCkRlY2F5PTUwLjAwCiNjbT1JbnAKSW5wdXQ9MC4wMApEcml2",
        "ZT0wLjAwClRvbmU9NTAuMDAKI2NtPVJldgpUZW5zaW9uPTAuMDAKRGVjYXk9NTAuMDAKU2hha2VNPTAuMDAKRGVuc2l0eT0wLjAwCkxDRF9ubzE9OApMQ0Rfbm8yPTkK",
        "TENEX25vMz0xMApUeXBlPTIKQ291cGxlPTAKQnJpZ2h0PTAKI2NtPUZsdApmcmVxPTE1MC4wMApzdGVyZW89MC4wMApyZXNvPTAuMDAKYmxlbmQ9MC4wMAp0eXBlPTEK",
        "I2NtPU91dApNaXg9NTAuMDAKV2lkdGg9MTAwLjAwCkdhaW5NaXg9MC4wMApCeXBhc3M9MApvdXRjbHA9MAojY209QW1wTW9kClNyYz0wCkRlcHRoPTAuMDAKI2NtPUxN",
        "ZW0KTGF5TWVtMT0wCgoKCgovLyBTZWN0aW9uIGZvciB1Z2x5IGNvbXByZXNzZWQgYmluYXJ5IERhdGEKLy8gRE9OJ1QgVE9VQ0ggVEhJUwoKJCQkJDEyOTgKP2dkZ2Nl",
        "bmViOmFhYWJhYWFhOmVhYWFhYWFhOmFrYWFhYWFhOmFhYWFpYWRwOmdqZ29namhlOmdqZ2JnbWdqOgpoa2dmY29naTpkY2hhYWFhYTpvYmNqcGNkbDphYTpjbjprZDpt",
        "bTphZTpjajpkYjpkYzppYTppaTppcDphYjphCmQ6YWc6YWk6YWs6YW46YW86Ymc6Y2w6ZGQ6ZGg6ZGk6ZGo6ZGs6ZHAhZ2RnY2VuZWJBTEEyNjVnZGdjZW5lYmUKYUEx",
        "MGdqZ29namhlZ2pnYmdtZ2poa2dmY29naUhoYUE0OWdkZ2NlbmViZWFBNzRnZGdjZW5lYmVhQTEwb2JGcApjZGxSaGRpb2ZhVGxqaW5nYmdhSmtla25tY01sYXBtZmlm",
        "am9rZm9obGhibG9GQ1FEb2Jua2trZWptYWdrR0toCmZtaG9wRkRtZktscGlkbGhFbG5ET25oQ2lma2FoZVNwb1ZZbGNKZ2RnY2VuZWJFQTEwTENOQWdkZ2NlbmViQUwK",
        "QTI2NWdkZ2NlbmViSUExMzhnZGdjZW5lYkFMQTI2NWdkZ2NlbmViUEE4SVplZEJIR0E1Z2RnY2VuZWJQQThJWgplZkJVV0E1Z2RnY2VuZWJQQThJWmVlQkhYQTU9NDM2",
        "NDMKAABdBgAAAAAAAFkGAAAjcGdtPWluaXRpYWxpemUuaDJwCiNBTT1Ud2FuZ3N0cm9tCiNWZXJzPTEwMDAwCiNFbmRpYW49bGl0dGxlCiNubT03CiNtcz1ub25lCiNt",
        "cz1Nb2RXaGwKI21zPVBpdGNoVwojbXM9Q3RybEEKI21zPUN0cmxCCiNtcz1MRk8KI21zPUVudmVsb3BlCiNudj0wCiNjbT1tYWluCkNyY1I9MQpDcmNSMj0xCiNjbT1Q",
        "Q29yZQpTVHJpZz0xClBQcm9qPTAKUEZvbGQ9MApQRmlsZT0xCkdGaWxlPTIKR1NjYWxlPTAKQ2hMYXk9MApTdXJyTz0xClJldj05Nzc1CkxFRD0wLjAwClBBR0U9MApQ",
        "YWdlc09uPTAKQ29yZU49MwpTbGljZT00ClVJX29wPTMKTWlkaUE9NQpNaWRpUD02CkRGb2xkPTcKQ3RybEE9MgpDdHJsQj0xMQojY209TU0xCkFjdGl2ZT0xClNvdXJj",
        "ZT0wCkRlc3QxPW5vbmU6YXNzaWduZWQKRGVwdGgxPTAuMDAKQ3VydmUxPTIKUW50emUxPTAKUmVjdDE9MApTSDE9MApTbGV3MT0wCiNjbT1NTTIKQWN0aXZlPTEKU291",
        "cmNlPTAKRGVzdDE9bm9uZTphc3NpZ25lZApEZXB0aDE9MC4wMApDdXJ2ZTE9MgpRbnR6ZTE9MApSZWN0MT0wClNIMT0wClNsZXcxPTAKI2NtPU1NMwpBY3RpdmU9MQpT",
        "b3VyY2U9MApEZXN0MT1ub25lOmFzc2lnbmVkCkRlcHRoMT0wLjAwCkN1cnZlMT0yClFudHplMT0wClJlY3QxPTAKU0gxPTAKU2xldzE9MAojY209TEZPClN5bmM9NgpX",
        "YXZlPTEKUmF0ZT0wLjAwCiNjbT1FbnYKU291cmNlPTAKTW9kZT0wClRocmVzPS0zMC4wMApBdHRhY2s9MjUuMDAKRGVjYXk9NTAuMDAKI2NtPUlucApJbnB1dD0wLjAw",
        "CkRyaXZlPTAuMDAKVG9uZT01MC4wMAojY209UmV2ClRlbnNpb249MC4wMApEZWNheT01MC4wMApTaGFrZU09MC4wMApEZW5zaXR5PTAuMDAKTENEX25vMT04CkxDRF9u",
        "bzI9OQpMQ0Rfbm8zPTEwClR5cGU9MgpDb3VwbGU9MApCcmlnaHQ9MAojY209Rmx0CmZyZXE9MTUwLjAwCnN0ZXJlbz0wLjAwCnJlc289MC4wMApibGVuZD0wLjAwCnR5",
        "cGU9MQojY209T3V0Ck1peD01MC4wMApXaWR0aD0xMDAuMDAKR2Fpbk1peD0wLjAwCkJ5cGFzcz0wCm91dGNscD0wCiNjbT1BbXBNb2QKU3JjPTAKRGVwdGg9MC4wMAoj",
        "Y209TE1lbQpMYXlNZW0xPTAKCgoKCi8vIFNlY3Rpb24gZm9yIHVnbHkgY29tcHJlc3NlZCBiaW5hcnkgRGF0YQovLyBET04nVCBUT1VDSCBUSElTCgokJCQkMTI5OAo/",
        "Z2RnY2VuZWI6YWFhYmFhYWE6ZWFhYWFhYWE6YWthYWFhYWE6YWFhYWlhZHA6Z2pnb2dqaGU6Z2pnYmdtZ2o6CmhrZ2Zjb2dpOmRjaGFhYWFhOm9iY2pwY2RsOmFhOmNu",
        "OmtkOm1tOmFlOmNqOmRiOmRjOmlhOmlpOmlwOmFiOmEKZDphZzphaTphazphbjphbzpiZzpjbDpkZDpkaDpkaTpkajpkazpkcCFnZGdjZW5lYkFMQTI2NWdkZ2NlbmVi",
        "ZQphQTEwZ2pnb2dqaGVnamdiZ21namhrZ2Zjb2dpSGhhQTQ5Z2RnY2VuZWJlYUE3NGdkZ2NlbmViZWFBMTBvYkZwCmNkbFJoZGlvZmFUbGppbmdiZ2FKa2Vrbm1jTWxh",
        "cG1maWZqb2tmb2hsaGJsb0ZDUURvYm5ra2tlam1hZ2tHS2gKZm1ob3BGRG1mS2xwaWRsaEVsbkRPbmhDaWZrYWhlU3BvVllsY0pnZGdjZW5lYkVBMTBMQ05BZ2RnY2Vu",
        "ZWJBTApBMjY1Z2RnY2VuZWJJQTEzOGdkZ2NlbmViQUxBMjY1Z2RnY2VuZWJQQThJWmVkQkhHQTVnZGdjZW5lYlBBOElaCmVmQlVXQTVnZGdjZW5lYlBBOElaZWVCSFhB",
        "NT00MzY0MwoAAA==",
        "AHByb2dyYW06IFByb2dyYW0gMAAQAAAA",
    ],
    {
        "drive": "15:15",
        "tone": "16:16",
        "tension": "17:17",
        "decay": "18:18",
        # "shake": "19:19",
        "density": "20:20",
    },
)



"""Generic utility functions."""

import bisect
import functools
import importlib
import itertools
import typing
import warnings

from mutwo.utilities import constants


__all__ = (
    "scale",
    "accumulate_from_n",
    "accumulate_from_zero",
    "insert_next_to",
    "uniqify_iterable",
    "cyclic_permutations",
    # "import_module_if_dependencies_have_been_installed",  # not for public use
    "find_closest_index",
    "find_closest_item",
    # "class_name_to_object_name",  # not for public use
    "get_nested_item_from_indices",
    "set_nested_item_from_indices",
)


def scale(
    value: constants.Real,
    old_min: constants.Real,
    old_max: constants.Real,
    new_min: constants.Real,
    new_max: constants.Real,
) -> constants.Real:
    """Scale a value from one range to another range.

    :param value: The value that shall be scaled.
    :param old_min: The minima of the old range.
    :param old_max: The maxima of the old range.
    :param new_min: The minima of the new range.
    :param new_max: The maxima of the new range.

    **Example:**

    >>> from mutwo.utilities import tools
    >>> tools.scale(1, 0, 1, 0, 100)
    100
    >>> tools.scale(0.5, 0, 1, 0, 100)
    50
    >>> tools.scale(0.2, 0, 1, 0, 100)
    20
    """

    try:
        assert old_min <= value <= old_max
    except AssertionError:
        msg = (
            "Input value '{}' has to be in the range of (old_min = {}, old_max = {})."
            .format(value, old_min, old_max)
        )
        raise ValueError(msg)
    return (((value - old_min) / (old_max - old_min)) * (new_max - new_min)) + new_min


def accumulate_from_n(
    iterable: typing.Iterable[constants.Real], n: constants.Real
) -> typing.Iterator:
    """Accumulates iterable starting with value n.

    :param iterable: The iterable which values shall be accumulated.
    :param n: The start number from which shall be accumulated.

    **Example:**

    >>> from mutwo.utilities import tools
    >>> tools.accumulate_from_n((4, 2, 3), 0)
    (0, 4, 6, 9)
    >>> tools.accumulate_from_n((4, 2, 3), 2)
    (2, 6, 8, 11)
    """
    return itertools.accumulate(itertools.chain((n,), iterable))


def accumulate_from_zero(iterable: typing.Iterable[constants.Real]) -> typing.Iterator:
    """Accumulates iterable starting from 0.

    :param iterable: The iterable which values shall be accumulated.

    **Example:**

    >>> from mutwo.utilities import tools
    >>> tools.accumulate_from_zero((4, 2, 3), 0)
    (0, 4, 6, 9)
    """
    return accumulate_from_n(iterable, 0)


def insert_next_to(
    iterable: typing.MutableSequence,
    item_to_find: typing.Any,
    distance: int,
    item_to_insert: typing.Any,
):
    """Insert an item into a list relative to the first item equal to a certain value."""

    index = list(iterable).index(item_to_find)
    if distance == 0:
        iterable[index] = item_to_insert
    else:
        real_distance = distance + 1 if distance < 0 else distance
        iterable.insert(index + real_distance, item_to_insert)


T = typing.TypeVar("T", bound=constants.Real)


def find_closest_index(
    item: constants.Real,
    data: typing.Iterable,
    key: typing.Callable[[typing.Any], T] = lambda item: item,
) -> int:
    """Return index of element in ``data`` with smallest difference to ``item``.

    :param item: The item from which the closest item shall be found.
    :param data: The data to which the closest item shall be found.

    **Example:**

    >>> from mutwo.utilities import tools
    >>> tools.find_closest_index(2, (1, 4, 5))
    0
    >>> tools.find_closest_index(127, (100, 4, 300, 53, 129))
    4
    >>> tools.find_closest_index(127, (('hi', 100), ('hey', 4), ('hello', 300)), key=lambda item: item[1])
    0
    """

    research_data = tuple(map(key, data))
    sorted_research_data = sorted(research_data)

    solution = bisect.bisect_left(sorted_research_data, item)
    # make type ignore because data has been converted to tuple (which is sizeable)
    if solution == len(data):  # type: ignore
        index = solution - 1

    elif solution == 0:
        index = solution

    else:
        indices = (solution, solution - 1)
        differences = tuple(abs(-sorted_research_data[n] + item) for n in indices)
        index = indices[differences.index(min(differences))]  # type: ignore

    return research_data.index(sorted_research_data[index])


def find_closest_item(
    item: constants.Real,
    data: typing.Sequence,
    key: typing.Callable[[typing.Any], T] = lambda item: item,
) -> T:
    """Return element in ``data`` with smallest difference to ``item``.

    :param item: The item from which the closest item shall be found.
    :param data: The data to which the closest item shall be found.
    :return: The closest number to ``item`` in ``data``.

    **Example:**

    >>> from mutwo.utilities import tools
    >>> tools.find_closest_item(2, (1, 4, 5))
    1
    >>> tools.find_closest_item(127, (100, 4, 300, 53, 129))
    129
    >>> tools.find_closest_item(127, (('hi', 100), ('hey', 4), ('hello', 300)), key=lambda item: item[1])
    ('hi', 100)
    """
    return data[find_closest_index(item, data, key=key)]


def import_module_if_dependencies_have_been_installed(
    module: str, dependencies: typing.Tuple[str, ...], import_class: bool = False
) -> None:
    for dependency in dependencies:
        try:
            importlib.import_module(dependency)
        except ImportError:
            message = (
                "Can't load module '{0}'. Install dependency '{1}' if you want to use"
                " '{0}'.".format(module, dependency)
            )
            warnings.warn(message)
            return

    if import_class:
        class_name = module.split(".")[-1]
        return getattr(importlib.import_module(module), class_name)

    else:
        importlib.import_module(module)


def uniqify_iterable(
    iterable: typing.Sequence,
    sort_key: typing.Callable[[typing.Any], constants.Real] = None,
    group_by_key: typing.Callable[[typing.Any], typing.Any] = None,
) -> typing.Iterable:
    """Not-Order preserving function to uniqify any iterable with non-hashable objects.

    :param iterable: The iterable which items shall be uniqified.
    :return: Return uniqified version of the entered iterable.
        The function will try to return the same type of the passed
        iterable. If Python raises an error during initialisation of
        the original iterable type, the function will simply return
        a tuple.

    **Example:**

    >>> from mutwo.parameters import pitches
    >>> from mutwo.utilities import tools
    >>> tools.uniqify_iterable([pitches.WesternPitch(pitch_name) for pitch_name in 'c d e c d e e f a c a'.split(' ')])
    [WesternPitch(c4),
    WesternPitch(d4),
    WesternPitch(e4),
    WesternPitch(f4),
    WesternPitch(a4)]
    """

    sorted_iterable = sorted(iterable, key=sort_key)
    result = (key for key, _ in itertools.groupby(sorted_iterable, key=group_by_key))

    try:
        return type(iterable)(result)  # type: ignore

    except Exception:
        return tuple(result)


def cyclic_permutations(sequence: typing.Sequence[typing.Any]) -> typing.Generator:
    """Cyclic permutation of an iterable. Return a generator object.

    :param sequence: The sequence from which cyclic permutations shall be generated.

    **Example:**

    >>> from mutwo.utilities import tools
    >>> permutations = tools.cyclic_permutations((1, 2, 3, 4))
    >>> next(permutations)
    (2, 3, 4, 1)
    >>> next(permutations)
    (3, 4, 1, 2)

    `Adapted function from the reply of Paritosh Singh
    <https://stackoverflow.com/questions/56171246/cyclic-permutation-operators-in-python/56171531>`_
    """

    def reorder_from_index(index: int, sequence: tuple) -> tuple:
        return sequence[index:] + sequence[:index]

    return (
        functools.partial(reorder_from_index, i)(sequence) for i in range(len(sequence))
    )


def class_name_to_object_name(class_name: str) -> str:
    """

    MyClassName -> my_class_name
    """
    characters = []

    is_first = True
    for character in class_name:
        if character.isupper():
            character = character.lower()
            if not is_first:
                character = "_{}".format(character)

        if is_first:
            is_first = False

        characters.append(character)

    return "".join(characters)


def get_nested_item_from_indices(
    indices: typing.Sequence[int], sequence: typing.Sequence
) -> typing.Any:
    """Get item in nested Sequence.

    :param indices: The indices of the nested item.
    :type indices: typing.Sequence[int]
    :param sequence: A nested sequence.
    :type sequence: typing.Sequence[typing.Any]


    **Example:**

    >>> from mutwo.utilities import tools
    >>> nested_sequence = (1, 2, (4, (5, 1), (9, (3,))))
    >>> tools.get_nested_item_from_indices((2, 2, 0), nested_sequence)
    9
    >>> nested_sequence[2][2][0]  # is equal
    9
    """

    for index in indices:
        sequence = sequence[index]
    return sequence


def set_nested_item_from_indices(
    indices: typing.Sequence[int], sequence: typing.MutableSequence, item: typing.Any,
) -> None:
    """Set item in nested Sequence.

    :param indices: The indices of the nested item which shall be set.
    :type indices: typing.Sequence[int]
    :param sequence: A nested sequence.
    :type sequence: typing.MutableSequence[typing.Any]
    :param item: The new item value.
    :type item: typing.Any

    **Example:**

    >>> from mutwo.utilities import tools
    >>> nested_sequence = [1, 2, [4, [5, 1], [9, [3]]]]]
    >>> tools.set_nested_item_from_indices((2, 2, 0), nested_sequence, 100)
    >>> nested_sequence[2][2][0] = 100  # is equal
    """

    n_indices = len(indices)
    for nth_index, index in enumerate(indices):
        if n_indices == nth_index + 1:
            sequence.__setitem__(index, item)
        else:
            sequence = sequence[index]



"""Build Abjad attachments from Mutwo data.
"""

import abc
import dataclasses
import typing
import warnings

import abjad  # type: ignore

from mutwo.parameters import abc as parameters_abc
from mutwo.parameters import notation_indicators
from mutwo.parameters import playing_indicators
from mutwo.utilities import tools


class AbjadAttachment(abc.ABC):
    """Abstract base class for all Abjad attachments."""

    @classmethod
    def get_class_name(cls):
        return tools.class_name_to_object_name(cls.__name__)

    @classmethod
    def from_indicator_collection(
        cls, indicator_collection: parameters_abc.IndicatorCollection
    ) -> typing.Optional["AbjadAttachment"]:
        """Initialize :class:`AbjadAttachment` from :class:`~mutwo.parameters.abc.IndicatorCollection`.

        If no suitable :class:`~mutwo.parameters.abc.Indicator` could be found in the collection
        the method will simply return None.
        """

        class_name = cls.get_class_name()
        try:
            indicator = getattr(indicator_collection, class_name)
        except AttributeError:
            return None

        return cls(**indicator.get_arguments_dict())

    @abc.abstractmethod
    def process_leaves(
        self,
        leaves: typing.Tuple[abjad.Leaf, ...],
        previous_attachment: typing.Optional["AbjadAttachment"],
    ) -> typing.Tuple[abjad.Leaf, ...]:
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def is_active(self) -> bool:
        raise NotImplementedError()


class ToggleAttachment(AbjadAttachment):
    """Abstract base class for Abjad attachments which behave like a toggle.

    In Western notation one can differentiate between elements which only get
    notated if they change (for instance dynamics, tempo) and elements which
    have to be notated again and again (for instance arpeggi or tremolo).
    Attachments that inherit from :class:`ToggleAttachment` represent elements
    which only get notated if their value changes.
    """

    @abc.abstractmethod
    def process_leaf(
        self, leaf: abjad.Leaf, previous_attachment: typing.Optional["AbjadAttachment"]
    ) -> abjad.Leaf:
        raise NotImplementedError()

    def process_leaves(
        self,
        leaves: typing.Tuple[abjad.Leaf, ...],
        previous_attachment: typing.Optional["AbjadAttachment"],
    ) -> typing.Tuple[abjad.Leaf, ...]:
        if previous_attachment != self:
            return (self.process_leaf(leaves[0], previous_attachment),) + leaves[1:]
        else:
            return leaves


class BangAttachment(AbjadAttachment):
    """Abstract base class for Abjad attachments which behave like a bang.

    In Western notation one can differentiate between elements which only get
    notated if they change (for instance dynamics, tempo) and elements which
    have to be notated again and again to be effective (for instance arpeggi or
    tremolo). Attachments that inherit from :class:`BangAttachment` represent
    elements which have to be notated again and again to be effective.
    """

    @abc.abstractmethod
    def process_first_leaf(self, leaf: abjad.Leaf) -> abjad.Leaf:
        raise NotImplementedError()

    @abc.abstractmethod
    def process_last_leaf(self, leaf: abjad.Leaf) -> abjad.Leaf:
        raise NotImplementedError()

    @abc.abstractmethod
    def process_central_leaf(self, leaf: abjad.Leaf) -> abjad.Leaf:
        raise NotImplementedError()

    def process_leaves(
        self,
        leaves: typing.Tuple[abjad.Leaf, ...],
        previous_attachment: typing.Optional["AbjadAttachment"],
    ) -> typing.Tuple[abjad.Leaf, ...]:
        n_leaves = len(leaves)

        new_leaves = []

        if n_leaves > 0:
            new_leaves.append(self.process_first_leaf(leaves[0]))

        if n_leaves > 2:
            for leaf in leaves[1:-1]:
                new_leaves.append(self.process_central_leaf(leaf))

        if n_leaves > 1:
            new_leaves.append(self.process_last_leaf(leaves[-1]))

        return tuple(new_leaves)


class BangEachAttachment(BangAttachment):
    @abc.abstractmethod
    def process_leaf(self, leaf: abjad.Leaf) -> abjad.Leaf:
        raise NotImplementedError()

    def process_first_leaf(self, leaf: abjad.Leaf) -> abjad.Leaf:
        return self.process_leaf(leaf)

    def process_last_leaf(self, leaf: abjad.Leaf) -> abjad.Leaf:
        return self.process_leaf(leaf)

    def process_central_leaf(self, leaf: abjad.Leaf) -> abjad.Leaf:
        return self.process_leaf(leaf)


class BangFirstAttachment(BangAttachment):
    @abc.abstractmethod
    def process_leaf(self, leaf: abjad.Leaf) -> abjad.Leaf:
        raise NotImplementedError()

    def process_first_leaf(self, leaf: abjad.Leaf) -> abjad.Leaf:
        return self.process_leaf(leaf)

    def process_last_leaf(self, leaf: abjad.Leaf) -> abjad.Leaf:
        return leaf

    def process_central_leaf(self, leaf: abjad.Leaf) -> abjad.Leaf:
        return leaf


class BangLastAttachment(BangAttachment):
    @abc.abstractmethod
    def process_leaf(self, leaf: abjad.Leaf) -> abjad.Leaf:
        raise NotImplementedError()

    def process_first_leaf(self, leaf: abjad.Leaf) -> abjad.Leaf:
        return leaf

    def process_last_leaf(self, leaf: abjad.Leaf) -> abjad.Leaf:
        return self.process_leaf(leaf)

    def process_central_leaf(self, leaf: abjad.Leaf) -> abjad.Leaf:
        return leaf

    def process_leaves(
        self,
        leaves: typing.Tuple[abjad.Leaf, ...],
        previous_attachment: typing.Optional["AbjadAttachment"],
    ) -> typing.Tuple[abjad.Leaf, ...]:
        n_leaves = len(leaves)
        if n_leaves > 0:
            return leaves[:-1] + (self.process_last_leaf(leaves[-1]),)
        else:
            return leaves


class Arpeggio(playing_indicators.Arpeggio, BangFirstAttachment):
    def process_leaf(self, leaf: abjad.Leaf) -> abjad.Leaf:
        abjad.attach(abjad.Arpeggio(direction=self.direction), leaf)
        return leaf


class Articulation(playing_indicators.Articulation, BangEachAttachment):
    def process_leaf(self, leaf: abjad.Leaf) -> abjad.Leaf:
        abjad.attach(abjad.Articulation(self.name), leaf)
        return leaf


class Tremolo(playing_indicators.Tremolo, BangEachAttachment):
    def process_leaf(self, leaf: abjad.Leaf) -> abjad.Leaf:
        abjad.attach(
            abjad.StemTremolo(self.n_flags * (2 ** leaf.written_duration.flag_count)),
            leaf,
        )
        return leaf


class ArtificalHarmonic(playing_indicators.ArtificalHarmonic, BangEachAttachment):
    @staticmethod
    def _change_note_head_style(leaf: abjad.Chord) -> None:
        abjad.tweak(leaf.note_heads[1]).NoteHead.style = "#'harmonic"

    @staticmethod
    def _convert_and_test_leaf(leaf: abjad.Leaf) -> typing.Tuple[abjad.Leaf, bool]:
        # return True if artifical_harmonic can be attached and False if
        # artifical harmonic can't be attached

        if isinstance(leaf, abjad.Chord):
            try:
                assert len(leaf.note_heads) == 1
            except AssertionError:
                message = (
                    "Can't attach artifical harmonic on chord with more or less"
                    " than one pitch!"
                )
                warnings.warn(message)
                return leaf, False

            return leaf, True

        elif isinstance(leaf, abjad.Note):
            new_abjad_leaf = abjad.Chord([leaf.written_pitch], leaf.written_duration,)
            for indicator in abjad.get.indicators(leaf):
                if type(indicator) != dict:
                    abjad.attach(indicator, new_abjad_leaf)

            return new_abjad_leaf, True

        else:
            message = (
                "Can't attach artifical harmonic on abjad leaf '{}' of type '{}'!"
                .format(leaf, type(leaf))
            )
            warnings.warn(message)
            return leaf, False

    def _get_second_pitch(self, abjad_pitch: abjad.Pitch) -> abjad.Pitch:
        return abjad_pitch + self.n_semitones

    def process_leaf(self, leaf: abjad.Leaf) -> abjad.Leaf:
        leaf, is_attachable = ArtificalHarmonic._convert_and_test_leaf(leaf)
        if is_attachable:
            first_pitch = leaf.note_heads[0].written_pitch
            second_pitch = self._get_second_pitch(first_pitch)
            leaf.written_pitches = abjad.PitchSegment([first_pitch, second_pitch])
            ArtificalHarmonic._change_note_head_style(leaf)
        return leaf


class StringContactPoint(playing_indicators.StringContactPoint, ToggleAttachment):
    def process_leaf(
        self, leaf: abjad.Leaf, previous_attachment: typing.Optional["AbjadAttachment"]
    ) -> abjad.Leaf:
        string_contact_point_markup = abjad.StringContactPoint(
            self.contact_point
        ).markup
        if previous_attachment:
            if previous_attachment.contact_point == "pizzicato":  # type: ignore
                string_contact_point_markup = abjad.Markup(
                    [
                        abjad.markups.MarkupCommand(
                            "caps",
                            "arco {}".format(
                                " ".join(
                                    string_contact_point_markup.contents[0].arguments
                                )
                            ),
                        )
                    ]
                )

        abjad.attach(
            abjad.Markup(
                [abjad.markups.MarkupCommand("fontsize", -2.4)]
                + string_contact_point_markup.contents,
                direction="up",
            ),
            leaf,
        )
        return leaf

    def process_leaves(
        self,
        leaves: typing.Tuple[abjad.Leaf, ...],
        previous_attachment: typing.Optional["AbjadAttachment"],
    ) -> typing.Tuple[abjad.Leaf, ...]:
        # don't attach ordinario at start (this is the default playing technique)
        if previous_attachment is not None or self.contact_point != "ordinario":
            return super().process_leaves(leaves, previous_attachment)
        else:
            return leaves


class Pedal(playing_indicators.Pedal, ToggleAttachment):
    def process_leaf(
        self, leaf: abjad.Leaf, previous_attachment: typing.Optional["AbjadAttachment"]
    ) -> abjad.Leaf:
        if self.pedal_activity:
            pedal_class = abjad.StartPianoPedal
        else:
            pedal_class = abjad.StopPianoPedal

        abjad.attach(pedal_class(self.pedal_type), leaf)
        return leaf

    def process_leaves(
        self,
        leaves: typing.Tuple[abjad.Leaf, ...],
        previous_attachment: typing.Optional["AbjadAttachment"],
    ) -> typing.Tuple[abjad.Leaf, ...]:
        # don't attach pedal down at start
        if previous_attachment is not None or self.is_active:
            return super().process_leaves(leaves, previous_attachment)
        else:
            return leaves


class Hairpin(playing_indicators.Hairpin, ToggleAttachment):
    def process_leaf(
        self, leaf: abjad.Leaf, previous_attachment: typing.Optional["AbjadAttachment"]
    ) -> typing.Tuple[abjad.Leaf, ...]:
        if self.symbol == "!":
            abjad.attach(
                abjad.StopHairpin(), leaf,
            )
        else:
            abjad.attach(
                abjad.StartHairpin(self.symbol), leaf,
            )
        return leaf


class BartokPizzicato(parameters_abc.ExplicitPlayingIndicator, BangFirstAttachment):
    def process_leaf(self, leaf: abjad.Leaf) -> abjad.Leaf:
        abjad.attach(
            abjad.LilyPondLiteral("\\snappizzicato", format_slot="absolute_after"),
            leaf,
        )
        return leaf


class Fermata(playing_indicators.Fermata, BangFirstAttachment):
    def process_leaf(self, leaf: abjad.Leaf) -> abjad.Leaf:
        abjad.attach(
            abjad.Fermata(self.fermata_type), leaf,
        )
        return leaf


class NaturalHarmonic(parameters_abc.ExplicitPlayingIndicator, BangFirstAttachment):
    def process_leaf(self, leaf: abjad.Leaf) -> abjad.Leaf:
        abjad.attach(
            abjad.LilyPondLiteral("\\flageolet", directed="up", format_slot="after"),
            leaf,
        )

        return leaf


class Prall(parameters_abc.ExplicitPlayingIndicator, BangFirstAttachment):
    def process_leaf(self, leaf: abjad.Leaf) -> abjad.Leaf:
        abjad.attach(
            abjad.LilyPondLiteral("^\\prall", format_slot="after"), leaf,
        )
        return leaf


class Tie(parameters_abc.ExplicitPlayingIndicator, BangLastAttachment):
    def process_leaf(self, leaf: abjad.Leaf) -> abjad.Leaf:
        abjad.attach(
            abjad.Tie(), leaf,
        )
        return leaf


class LaissezVibrer(parameters_abc.ExplicitPlayingIndicator, BangLastAttachment):
    def process_leaf(self, leaf: abjad.Leaf) -> abjad.Leaf:
        abjad.attach(
            abjad.LaissezVibrer(), leaf,
        )
        return leaf


class BarLine(notation_indicators.BarLine, BangLastAttachment):
    def process_leaf(self, leaf: abjad.Leaf) -> abjad.Leaf:
        abjad.attach(
            abjad.BarLine(self.abbreviation), leaf,
        )
        return leaf


class Clef(notation_indicators.Clef, BangFirstAttachment):
    def process_leaf(self, leaf: abjad.Leaf) -> abjad.Leaf:
        abjad.attach(
            abjad.Clef(self.name), leaf,
        )
        return leaf


class Ottava(notation_indicators.Ottava, ToggleAttachment):
    def process_leaf(
        self, leaf: abjad.Leaf, previous_attachment: typing.Optional["AbjadAttachment"]
    ) -> abjad.Leaf:
        abjad.attach(
            abjad.Ottava(self.n_octaves, format_slot="before"), leaf,
        )
        return leaf

    def process_leaves(
        self,
        leaves: typing.Tuple[abjad.Leaf, ...],
        previous_attachment: typing.Optional["AbjadAttachment"],
    ) -> typing.Tuple[abjad.Leaf, ...]:
        # don't attach ottava = 0 at start (this is the default notation)
        if previous_attachment is not None or self.n_octaves != 0:
            return super().process_leaves(leaves, previous_attachment)
        else:
            return leaves


class Markup(notation_indicators.Markup, BangFirstAttachment):
    def process_leaf(self, leaf: abjad.Leaf) -> abjad.Leaf:
        abjad.attach(
            abjad.Markup(self.content, direction=self.direction), leaf,
        )
        return leaf


class RehearsalMark(notation_indicators.RehearsalMark, BangFirstAttachment):
    def process_leaf(self, leaf: abjad.Leaf) -> abjad.Leaf:
        abjad.attach(
            abjad.RehearsalMark(markup=self.markup), leaf,
        )
        return leaf


class MarginMarkup(notation_indicators.MarginMarkup, BangFirstAttachment):
    def process_leaf(self, leaf: abjad.Leaf) -> abjad.Leaf:
        command = "\\set {}.instrumentName = \\markup ".format(self.context)
        command += "{ " + self.content + " }"  # type: ignore
        abjad.attach(
            abjad.LilyPondLiteral(command), leaf,
        )
        return leaf


class Ornamentation(playing_indicators.Ornamentation, BangFirstAttachment):
    _direction_to_ornamentation_command = {
        "up": """
    #'((moveto 0 0)
      (lineto 0.5 0)
      (curveto 0.5 0 1.5 1.75 2.5 0)
      (lineto 3.5 0))""",
        "down": """
    #'((moveto 0 0)
      (lineto 0.5 0)
      (curveto 0.5 0 1.5 -1.75 2.5 0)
      (lineto 3.5 0))""",
    }

    def _make_markup(self) -> abjad.Markup:
        return abjad.Markup(
            [
                abjad.markups.MarkupCommand("vspace", -0.25),
                abjad.markups.MarkupCommand("fontsize", -4),
                abjad.markups.MarkupCommand("rounded-box", ["{}".format(self.n_times)]),
                abjad.markups.MarkupCommand("hspace", -0.4),
                abjad.markups.MarkupCommand(
                    "path",
                    0.25,
                    abjad.LilyPondLiteral(
                        self._direction_to_ornamentation_command[self.direction]  # type: ignore
                    ),
                ),
            ],
            direction="up",
        )

    def process_leaf(self, leaf: abjad.Leaf) -> abjad.Leaf:
        abjad.attach(self._make_markup(), leaf)
        return leaf


@dataclasses.dataclass()
class Dynamic(ToggleAttachment):
    dynamic_indicator: str = "mf"  # TODO(for future usage add typing.Literal)

    @classmethod
    def from_indicator_collection(
        cls, indicator_collection: parameters_abc.IndicatorCollection
    ) -> typing.Optional["AbjadAttachment"]:
        """Always return None.

        Dynamic can't be initialised from IndicatorCollection.
        """
        return None

    @property
    def is_active(self) -> bool:
        return True

    def process_leaf(
        self, leaf: abjad.Leaf, previous_attachment: typing.Optional["AbjadAttachment"]
    ) -> abjad.Leaf:
        abjad.attach(abjad.Dynamic(self.dynamic_indicator), leaf)
        return leaf


@dataclasses.dataclass()
class Tempo(BangFirstAttachment):
    reference_duration: typing.Optional[typing.Tuple[int, int]] = (1, 4)
    units_per_minute: typing.Optional[int] = 60
    textual_indication: typing.Optional[str] = None
    # TODO(for future usage add typing.Literal['rit.', 'acc.'])
    dynamic_change_indication: typing.Optional[str] = None
    stop_dynamic_change_indicaton: bool = False
    print_metronome_mark: bool = True

    @classmethod
    def from_indicator_collection(
        cls, indicator_collection: parameters_abc.IndicatorCollection
    ) -> typing.Optional["AbjadAttachment"]:
        """Always return None.

        Tempo can't be initialised from IndicatorCollection.
        """
        return None

    @property
    def is_active(self) -> bool:
        return True

    def _attach_metronome_mark(self, leaf: abjad.Leaf) -> None:
        if self.print_metronome_mark:
            abjad.attach(
                abjad.MetronomeMark(
                    reference_duration=self.reference_duration,
                    units_per_minute=self.units_per_minute,
                    textual_indication=self.textual_indication,
                ),
                leaf,
            )

    def process_leaf(self, leaf: abjad.Leaf) -> abjad.Leaf:
        self._attach_metronome_mark(leaf)

        if self.dynamic_change_indication is not None:
            dynamic_change_indication = abjad.StartTextSpan(
                left_text=abjad.Markup(self.dynamic_change_indication)
            )
            abjad.attach(dynamic_change_indication, leaf)

        return leaf


class DynamicChangeIndicationStop(BangFirstAttachment):
    @classmethod
    def from_indicator_collection(
        cls, indicator_collection: parameters_abc.IndicatorCollection
    ) -> typing.Optional["AbjadAttachment"]:
        """Always return None.

        DynamicChangeIndicationStop can't be initialised from IndicatorCollection.
        """
        return None

    @property
    def is_active(self) -> bool:
        return True

    def process_leaf(self, leaf: abjad.Leaf) -> abjad.Leaf:
        abjad.attach(abjad.StopTextSpan(), leaf)
        return leaf



"""Event classes which are designated for musical usage."""

try:
    import quicktions as fractions  # type: ignore
except ImportError:
    import fractions  # type: ignore

import numbers
import typing

from mutwo import events
from mutwo import parameters
from mutwo.utilities import constants

__all__ = ("NoteLike",)

PitchOrPitches = typing.Union[
    parameters.abc.Pitch, typing.Iterable, constants.Real, None
]

Volume = typing.Union[parameters.abc.Volume, constants.Real, str]


class NoteLike(events.basic.SimpleEvent):
    """NoteLike represents traditional discreet musical objects.

    :param pitch_or_pitches: The pitch or pitches of the event. This can
        be a pitch object (any class that inherits from ``mutwo.parameters.abc.Pitch``)
        or a list of pitch objects. Furthermore mutwo supports syntactic sugar
        to convert other objects on the fly to pitch objects: Atring can be
        read as pitch class names to build
        :class:`mutwo.parameters.pitches.WesternPitch` objects or as ratios to
        build :class:`mutwo.parameters.pitches.JustIntonationPitch` objects.
        Fraction will also build :class:`mutwo.parameters.pitches.JustIntonationPitch`
        objects. Other numbers (integer and float) will be read as pitch class numbers
        to make :class:`mutwo.parameters.pitches.WesternPitch` objects.
    :param new_duration: The duration of ``NoteLike``. This can be any number.
        The unit of the duration is up to the interpretation of the user and the
        respective converter routine that will be used.
    :param volume: The volume of the event. Can either be a object of
        :mod:`mutwo.parameters.volumes`, a number or a string. If the number
        ranges from 0 to 1, mutwo automatically generates a
        :class:`mutwo.parameters.volumes.DirectVolume` object (and the number
        will be interpreted as the amplitude). If the
        number is smaller than 0, automatically generates a
        :class:`mutwo.parameters.volumes.DecibelVolume` object (and the number
        will be interpreted as decibel). If the argument is a string,
        `mutwo` will try to initialise a :class:`mutwo.parameters.volumes.WesternVolume`
        object.
    :param playing_indicators:
    :param notation_indicators:

    By default mutwo doesn't differentiate between Tones, Chords and
    Rests, but rather simply implements one general class which can
    represent any of the mentioned definitions (e.g. a NoteLike object
    with several pitches may be called a 'Chord' and a NoteLike object
    with only one pitch may be called a 'Tone').

    **Example:**

    >>> from mutwo.parameters import pitches
    >>> from mutwo.events import music
    >>> tone = music.NoteLike(pitches.WesternPitch('a'), 1, 1)
    >>> other_tone = music.NoteLike('3/2', 1, 0.5)
    >>> chord = music.NoteLike(
        [pitches.WesternPitch('a'), pitches.JustIntonationPitch('3/2')], 1, 1
    )
    >>> other_chord = music.NoteLike('c4 dqs3 10/7', 1, 3)
    """

    def __init__(
        self,
        pitch_or_pitches: PitchOrPitches = "c",
        duration: parameters.abc.DurationType = 1,
        volume: Volume = "mf",
        playing_indicators: parameters.playing_indicators.PlayingIndicatorCollection = None,
        notation_indicators: parameters.notation_indicators.NotationIndicatorCollection = None,
        # before_grace_notes  # TODO(add grace note container!)
        # after_grace_notes
    ):
        if playing_indicators is None:
            playing_indicators = (
                events.music_constants.DEFAULT_PLAYING_INDICATORS_COLLECTION_CLASS()
            )

        if notation_indicators is None:
            notation_indicators = (
                events.music_constants.DEFAULT_NOTATION_INDICATORS_COLLECTION_CLASS()
            )

        self.pitch_or_pitches = pitch_or_pitches
        self.volume = volume
        super().__init__(duration)
        self.playing_indicators = playing_indicators
        self.notation_indicators = notation_indicators

    # ###################################################################### #
    #                          static methods                                #
    # ###################################################################### #

    @staticmethod
    def _convert_string_to_pitch(pitch_indication: str) -> parameters.abc.Pitch:
        # assumes it is a ratio
        if "/" in pitch_indication:
            return parameters.pitches.JustIntonationPitch(pitch_indication)

        # assumes it is a WesternPitch name
        elif (
            pitch_indication[0]
            in parameters.pitches_constants.DIATONIC_PITCH_NAME_TO_PITCH_CLASS.keys()
        ):
            if pitch_indication[-1].isdigit():
                pitch_name, octave = pitch_indication[:-1], int(pitch_indication[-1])
                pitch = parameters.pitches.WesternPitch(pitch_name, octave)
            else:
                pitch = parameters.pitches.WesternPitch(pitch_indication)

            return pitch

        else:
            message = (
                "Can't build pitch from pitch_indication '{}'. Supported string formats"
                " are (1) ratios divided by a forward slash (for instance '3/2' or"
                " '4/3') and (2) names of western pitch classes with an optional number"
                " to indicate the octave (for instance 'c4', 'as' or 'fqs2')."
            )
            raise NotImplementedError(message)

    @staticmethod
    def _convert_fraction_to_pitch(
        pitch_indication: fractions.Fraction,
    ) -> parameters.abc.Pitch:
        return parameters.pitches.JustIntonationPitch(pitch_indication)

    @staticmethod
    def _convert_float_or_integer_to_pitch(
        pitch_indication: float,
    ) -> parameters.abc.Pitch:
        return parameters.pitches.WesternPitch(pitch_indication)

    @staticmethod
    def _convert_unknown_object_to_pitch(
        unknown_object: typing.Any,
    ) -> typing.List[parameters.abc.Pitch]:
        if unknown_object is None:
            pitches = []

        elif isinstance(unknown_object, parameters.abc.Pitch):
            pitches = [unknown_object]

        elif isinstance(unknown_object, str):
            pitches = [
                NoteLike._convert_string_to_pitch(pitch_indication)
                for pitch_indication in unknown_object.split(" ")
            ]

        elif isinstance(unknown_object, fractions.Fraction):
            pitches = [NoteLike._convert_fraction_to_pitch(unknown_object)]

        elif isinstance(unknown_object, float) or isinstance(unknown_object, int):
            pitches = [NoteLike._convert_float_or_integer_to_pitch(unknown_object)]

        else:
            message = "Can't build pitch object from object '{}' of type '{}'.".format(
                unknown_object, type(unknown_object)
            )
            raise NotImplementedError(message)

        return pitches

    # ###################################################################### #
    #                            properties                                  #
    # ###################################################################### #

    @property
    def _parameters_to_print(self) -> typing.Tuple[str, ...]:
        """Return tuple of attribute names which shall be printed for repr.
        """
        return tuple(
            attribute
            for attribute in self._parameters_to_compare
            if attribute not in ("playing_indicators", "notation_indicators")
        )

    @property
    def pitch_or_pitches(self) -> typing.Any:
        """The pitch or pitches of the event."""

        return self._pitch_or_pitches

    @pitch_or_pitches.setter
    def pitch_or_pitches(self, pitch_or_pitches: typing.Any):
        # make sure pitch_or_pitches always become assigned to a list of pitches,
        # to be certain of the returned type
        if not isinstance(pitch_or_pitches, str) and isinstance(
            pitch_or_pitches, typing.Iterable
        ):
            # several pitches
            pitches_per_element = (
                NoteLike._convert_unknown_object_to_pitch(pitch)
                for pitch in pitch_or_pitches
            )
            pitch_or_pitches = []
            for pitches in pitches_per_element:
                pitch_or_pitches.extend(pitches)
        else:
            pitch_or_pitches = NoteLike._convert_unknown_object_to_pitch(
                pitch_or_pitches
            )

        self._pitch_or_pitches = pitch_or_pitches

    @property
    def volume(self) -> typing.Any:
        """The volume of the event."""

        return self._volume

    @volume.setter
    def volume(self, volume: typing.Any):
        if isinstance(volume, numbers.Real):
            if volume >= 0:  # type: ignore
                volume = parameters.volumes.DirectVolume(volume)  # type: ignore
            else:
                volume = parameters.volumes.DecibelVolume(volume)  # type: ignore

        elif isinstance(volume, str):
            volume = parameters.volumes.WesternVolume(volume)

        elif not isinstance(volume, parameters.abc.Volume):
            message = (
                "Can't initialise '{}' with value '{}' of type '{}' for argument"
                " 'volume'. The type for 'volume' should be '{}'.".format(
                    type(self).__name__, volume, type(volume), Volume
                )
            )
            raise TypeError(message)
        self._volume = volume



import typing

import rpp  # type: ignore

from mutwo.converters.frontends import reaper_constants

from mutwo import converters
from mutwo import events
from mutwo import parameters
from mutwo.utilities import tools

__all__ = ("ReaperFileConverter", "ReaperMarkerConverter")


class ReaperFileConverter(converters.abc.Converter):
    def __init__(self):
        self.reaper_project = self.create_reaper_project()

    @staticmethod
    def create_reaper_project() -> rpp.Element:
        return rpp.Element(
            tag="REAPER_PROJECT",
            attrib=["0.1", "6.16/linux-x86_64", "1607708341"],
            children=[],
        )

    def add_track(self, name: str = "") -> rpp.Element:
        track = rpp.Element(
            tag="TRACK",
            children=[["NAME", name], rpp.Element(tag="FXCHAIN", children=[],),],
        )
        self.reaper_project.append(track)
        return track

    def get_track(self, name: str) -> rpp.Element:
        for track in self.reaper_project.findall(".//TRACK"):
            current_trackname = track.find(".//NAME")
            if current_trackname[1] == name:
                return track

    @staticmethod
    def add_fx(
        track: rpp.Element, fx_info: reaper_constants.VstInfo, name: str = ""
    ) -> rpp.Element:
        fx = rpp.Element(
            tag="VST", attrib=[name, fx_info.filename], children=fx_info.binary
        )
        fxchain = track.find(".//FXCHAIN")
        fxchain.append(fx)
        return fx

    @staticmethod
    def get_all_fx(track) -> typing.List[rpp.Element]:
        fxchain = track.find(".//FXCHAIN")
        return fxchain.findall(".//VST")

    @staticmethod
    def add_parameter_envelope(
        track: rpp.Element,
        fx: rpp.Element,
        parameter_name: str,
        envelope_events: events.basic.SequentialEvent,
    ) -> rpp.Element:
        """events need a value attribute"""
        envelope = rpp.Element(
            tag="PARMENV",
            attrib=[parameter_name, "0", "1", "0.5"],
            children=[
                ["ACT", "1", "-1"],
                ["VIS", "1", "1", "1"],
                ["DEFSHAPE", "0", "-1", "-1"],
            ],
        )
        envelope_points = [
            ["PT", time, event.value, 0]
            for time, event in zip(envelope_events.absolute_times, envelope_events)
        ]
        envelope.extend(envelope_points)
        fxchain = track.find(".//FXCHAIN")
        tools.insert_next_to(fxchain, fx, 1, envelope)
        return envelope

    def convert(self, event: events.abc.Event):
        with open(self.path, "w") as file:
            rpp.dump(self.reaper_project, file)


class ReaperMarkerConverter(converters.abc.EventConverter):
    """Make Reaper Marker entries."""

    def __init__(
        self,
        simple_event_to_marker_name: typing.Callable[
            [events.basic.SimpleEvent], str
        ] = lambda simple_event: simple_event.name,  # type: ignore
        simple_event_to_marker_color: typing.Callable[
            [events.basic.SimpleEvent], str
        ] = lambda simple_event: simple_event.color,  # type: ignore
    ):
        self._simple_event_to_marker_name = simple_event_to_marker_name
        self._simple_event_to_marker_color = simple_event_to_marker_color

    def _convert_simple_event(
        self,
        simple_event: events.basic.SimpleEvent,
        absolute_entry_delay: parameters.abc.DurationType,
    ) -> typing.Tuple[str, ...]:
        try:
            marker_name = self._simple_event_to_marker_name(simple_event)
            marker_color = self._simple_event_to_marker_color(simple_event)
        except AttributeError:
            return tuple([])

        return ("{} {} {}".format(absolute_entry_delay, marker_name, marker_color),)

    def convert(self, event_to_convert: events.abc.Event) -> str:
        reaper_marker = tuple(
            "MARKER {} {}".format(nth_marker, marker_data)
            for nth_marker, marker_data in enumerate(
                self._convert_event(event_to_convert, 0)
            )
        )
        return "\n".join(reaper_marker)
